import chess.pgn
import pandas as pd
from pandas import DataFrame

from chipiron.environments.chess.board.utils import fen

# atm this file is not shared on GitHub so download one like this. This one was downloaded from the free online database of lichess
pgn = open(
    "/home/victor_old/pycharm/chipiron/data/lichess_db_standard_rated_2015-03.pgn"
)
data_frame_file_name = "/home/victor/good_games2025"
the_dic: list[dict[str, fen]] = []

count_game: int = 0
total_count_move: int = 0
recorded_board = 0
while recorded_board < 10000000:

    count_game += 1
    if count_game % 1000 == 0:
        print("count", count_game)
    game: chess.pgn.GameNode | None = chess.pgn.read_game(pgn)

    if count_game % 10000 == 0:  # save
        # weird that an array is given where a dictionary is expected
        new_data_frame_states: DataFrame = pd.DataFrame.from_dict(the_dic)
        print("%%", len(new_data_frame_states.index))
        recorded_board = len(new_data_frame_states.index)
        new_data_frame_states.to_pickle(data_frame_file_name)
    if game is None:
        print("GAME NONE")
        break
    else:
        # Iterate through all moves and play them on a board.
        chess_board = game.board()

        round_: int = 0
        for move in game.mainline_moves():
            total_count_move += 1
            round_ += 1
            #     try:
            chess_board.push(move)
            assert game is not None
            game = game.next()

            assert game is not None
            if game.eval() is not None and total_count_move % 50 == 0:
                the_dic.append({"fen": chess_board.fen()})



"""
Module to define the messages that are sent between the players and the game.
"""

from dataclasses import dataclass

import chess

from chipiron.environments.chess.board.utils import FenPlusHistory, fen
from chipiron.environments.chess.move import moveUci
from chipiron.players.boardevaluators.board_evaluation.board_evaluation import (
    BoardEvaluation,
)
from chipiron.players.player import PlayerId


@dataclass
class MoveMessage:
    """
    Represents a message containing a move made by a player.

    Attributes:
        move (chess.Move): The move made by the player.
        corresponding_board (fen): The FEN representation of the board after the move.
        player_name (str): The name of the player who made the move.
        color_to_play (chess.Color): The color of the player to play after the move.
        evaluation (float | None, optional): The evaluation score of the move. Defaults to None.
    """

    move: moveUci
    corresponding_board: fen
    player_name: PlayerId
    color_to_play: chess.Color
    evaluation: BoardEvaluation | None = None


@dataclass
class BoardMessage:
    """
    Represents a message containing the current state of the chess board.

    Attributes:
        fen_plus_moves (FenPlusMoves): The original fen and the subsequent moves to be applied.
        seed (int | None, optional): The seed used for random number generation. Defaults to None.
    """

    fen_plus_moves: FenPlusHistory
    seed: int | None = None



from typing import Any
from unittest.mock import MagicMock

import chess

from chipiron.environments.chess.board import IBoard
from chipiron.players.move_selector.treevalue.nodes.tree_node import TreeNode


class TestTreeNode:
    def setup_method(self) -> None:
        print("Setting up the test for the TreeNode class.")
        self.mock_board = MagicMock(spec=IBoard)
        self.mock_board.legal_moves = MagicMock()
        self.mock_board.legal_moves.get_all = MagicMock(return_value=["e2e4", "d2d4"])
        self.mock_board.turn = chess.WHITE
        self.tree_node: TreeNode[Any] = TreeNode(
            id_=1,
            half_move_=0,
            board_=self.mock_board,
            parent_nodes_={},
            all_legal_moves_generated=False,
            non_opened_legal_moves=set(),
            moves_children_={},
        )

    def test_all_legal_moves_generated_true(self) -> None:
        self.tree_node.all_legal_moves_generated = True
        self.tree_node.moves_children_ = {0: None}
        self.tree_node.non_opened_legal_moves = {0}
        self.tree_node.test_all_legal_moves_generated()

    def test_all_legal_moves_generated_false(self) -> None:
        self.tree_node.all_legal_moves_generated = False
        self.tree_node.moves_children_ = {0: None}
        self.tree_node.test_all_legal_moves_generated()

    def test_all_legal_moves_generated_no_moves(self) -> None:
        self.tree_node.all_legal_moves_generated = False
        self.mock_board.legal_moves.get_all = MagicMock(return_value=[])
        self.tree_node.test_all_legal_moves_generated()


if __name__ == "__main__":
    test = TestTreeNode()
    test.setup_method()
    test.test_all_legal_moves_generated_true()
    test.test_all_legal_moves_generated_false()
    test.test_all_legal_moves_generated_no_moves()
    print("All tests passed!")



"""
Launching the main chipiron
"""

import argparse
import sys
from typing import Any

from chipiron.scripts.factory import create_script
from chipiron.scripts.iscript import IScript
from chipiron.scripts.script_gui.script_gui_custom import script_gui
from chipiron.scripts.script_type import ScriptType

sys.path.append("../../")


def get_script_and_args(
    raw_command_line_arguments: list[str],
) -> tuple[ScriptType, dict[str, Any] | None, str | None]:
    """

    Args:
        raw_command_line_arguments: the list of arguments of the scripts given by command line

    Returns:
        A string for the name of script and a dictionary of parameters

    """
    script_type: ScriptType
    extra_args: dict[str, Any] | None = None
    config_file_name: str | None = None
    # Whether command line arguments are provided or not we ask for more info through a GUI
    if len(raw_command_line_arguments) == 1:  # No args provided
        # use a gui to get user input
        gui_extra_args: dict[str, Any] | None
        script_type, gui_extra_args, config_file_name = script_gui()
        extra_args = gui_extra_args
    else:
        # first parse/retrieve the name of the script then look for the names of the parameters related to this script
        # then parse again and retrieve the parameters related to the script if specified

        # Capture  the script argument in the command line arguments
        parser_default: argparse.ArgumentParser = argparse.ArgumentParser()
        parser_default.add_argument(
            "--script_name", type=str, default=None, help="name of the script"
        )
        args_obj, _ = parser_default.parse_known_args()
        args_command_line: dict[Any, Any] = vars(
            args_obj
        )  # converting into dictionary format

        # print("command line arguments:", args_command_line)

        # the script name must be specified otherwise fail
        if args_command_line["script_name"] is None:
            raise ValueError(
                "Expecting command line arguments of the shape python chipiron.py --script_name **name_of script**"
            )

        script_type_str: str = args_command_line["script_name"]
        script_type = ScriptType(script_type_str)

        extra_args = {}

    # print("extra_args", extra_args)
    return script_type, extra_args, config_file_name


def main() -> None:
    """
    The main function
    """
    # Getting the command line arguments from the system
    raw_command_line_arguments: list[str] = sys.argv

    # the type of script to be executed
    script_type: ScriptType

    # arguments provided to the script from the outside. Here it can be from a gui or command line
    extra_args: dict[str, Any] | None

    # extracting the script_name and possibly some input arguments from either the gui or a yaml file or command line
    script_type, extra_args, config_file_name = get_script_and_args(
        raw_command_line_arguments
    )

    # creating the script object from its name and arguments
    script_object: IScript = create_script(
        script_type=script_type,
        extra_args=extra_args,
        config_file_name=config_file_name,
        should_parse_command_line_arguments=True,
    )

    # run the script
    script_object.run()

    # terminate the script
    script_object.terminate()


if __name__ == "__main__":
    # checking if the version of python is high enough
    message = (
        "A version of Python higher than 3.10 is required to run chipiron.\n"
        + ' Try using "python3 main_chipiron.py" instead'
    )

    assert sys.version_info >= (3, 10), message
    # launching the real main python script.
    # this allows the to bypass the automatic full interpreter check of python that would raise a syntax error before
    # the assertion above in case of a wrong python version
    main()



"""
This module defines the neural network model NetPP2D2_2_PRELU.

The NetPP2D2_2_PRELU class is a subclass of ChiNN and implements the forward pass of the neural network.
It consists of two fully connected layers with PReLU activation functions.


Methods:
    __init__(): Initializes the NetPP2D2_2_PRELU model.
    forward(x: torch.Tensor) -> torch.Tensor: Performs the forward pass of the neural network.
    init_weights(file: str) -> None: Initializes the weights of the model.
    print_param() -> None: Prints the parameters of the model.
    print_input(input: torch.Tensor) -> None: Prints the input tensor.

Helper Functions:
    print_piece_param(i: int, vec: torch.Tensor) -> None: Prints the parameters of a specific piece.

"""

from dataclasses import dataclass
from typing import Any, Callable, Dict, Literal

import torch
import torch.nn as nn
import yaml

from chipiron.players.boardevaluators.neural_networks.NNModelType import (
    ActivationFunctionType,
    NNModelType,
    activation_map,
)
from chipiron.utils.chi_nn import ChiNN
from chipiron.utils.logger import chipiron_logger


@dataclass
class MultiLayerPerceptronArgs:
    type: Literal[NNModelType.MultiLayerPerceptron]
    number_neurons_per_layer: list[int]
    list_of_activation_functions: list[ActivationFunctionType]

    def __str__(self) -> str:
        neurons = "-".join(map(str, self.number_neurons_per_layer))
        activations = "-".join(act.value for act in self.list_of_activation_functions)
        return f"Type: {self.type}, Neurons: [{neurons}], Activations: [{activations}]"

    def filename(self) -> str:
        # Create a filename-safe version by replacing spaces with underscores and ensuring everything is lowercase
        neurons = "_".join(map(str, self.number_neurons_per_layer))
        activations = "_".join(act.value for act in self.list_of_activation_functions)
        return f"{self.type}_{neurons}_{activations}"


def build_sequential(
    layers: list[int], activations: list[ActivationFunctionType]
) -> nn.Sequential:
    modules: list[nn.Module] = []
    for i in range(len(layers) - 1):
        modules.append(nn.Linear(layers[i], layers[i + 1]))
        act: ActivationFunctionType = activations[i]
        if act in activation_map:
            modules.append(activation_map[act]())
    return nn.Sequential(*modules)


def extract_sequential_model_data(model: nn.Sequential) -> Dict[str, Any]:
    """
    Extracts the weights and biases from a Sequential model in a layered dictionary format.

    Args:
        model (nn.Sequential): The Sequential model to extract from.

    Returns:
        Dict[str, Any]: A dictionary with layer names as keys and parameter data as values.
    """
    layers_data = {}

    for idx, layer in enumerate(model):
        layer_info = {}
        if hasattr(layer, "weight") and layer.weight is not None:
            layer_info["weight"] = layer.weight.detach().cpu().numpy().tolist()
        if hasattr(layer, "bias") and layer.bias is not None:
            layer_info["bias"] = layer.bias.detach().cpu().numpy().tolist()

        layer_name = f"layer_{idx}_{layer.__class__.__name__}"
        layers_data[layer_name] = layer_info if layer_info else "No parameters"

    return layers_data


class MultiLayerPerceptron(ChiNN):
    """
    Neural network model for board evaluation using 2D2_2_PRELU architecture.

    Attributes:
        fc1 (nn.Linear): The first fully connected layer.
        relu_1 (nn.PReLU): The first PReLU activation function.
        fc2 (nn.Linear): The second fully connected layer.
        tanh (nn.Tanh): The tanh activation function.
    """

    def __init__(self, args: MultiLayerPerceptronArgs) -> None:
        """Constructor for the NetPP2D2_2_PRELU class. Initializes the neural network layers."""
        super(MultiLayerPerceptron, self).__init__()

        self.model: Callable[[torch.Tensor], torch.Tensor] = build_sequential(
            layers=args.number_neurons_per_layer,
            activations=args.list_of_activation_functions,
        )

    def __getstate__(self) -> None:
        """Get the state of the neural network."""
        return None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the neural network.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor.
        """
        y = self.model(x)
        return y

    def init_weights(self) -> None:
        """Initialize the weights of the neural network."""
        pass

    def print_input(self, input: torch.Tensor) -> None:
        """
        Print the input tensor.

        Args:
            input (torch.Tensor): The input tensor.
        """

    ...

    def log_readable_model_weights_to_file(self, file_path: str) -> None:
        """
        Writes the model weights and biases into a YAML file.

        Args:
            file_path (str): The path where the YAML file will be saved.
        """
        assert isinstance(self.model, nn.Sequential)
        layers_data = extract_sequential_model_data(self.model)

        with open(file_path, "w") as f:
            yaml.dump(layers_data, f, default_flow_style=False, sort_keys=False)

        chipiron_logger.info(f"Model weights successfully written to {file_path}")

    def print_param(self) -> None:
        """
        Prints the model weights and biases to the console in YAML format.

        Args:
            model (nn.Sequential): The Sequential model.
        """
        assert isinstance(self.model, nn.Sequential)
        layers_data = extract_sequential_model_data(self.model)
        yaml_str = yaml.dump(layers_data, default_flow_style=False, sort_keys=False)
        chipiron_logger.info(yaml_str)


def print_piece_param(i: int, vec: torch.Tensor) -> None:
    """
    Prints the piece parameter at index i from the given tensor.

    Args:
        i (int): The index of the piece parameter to print.
        vec (torch.Tensor): The tensor containing the piece parameters.

    Returns:
        None
    """
    ...



import os
from enum import Enum

import parsley_coco

from chipiron.games.match.match_settings_args import MatchSettingsArgs
from chipiron.utils import path


class MatchConfigTag(str, Enum):
    # this list should correspond to files existing in the data/settings/OneMatch folder

    Cubo = "setting_cubo"
    Duda = "setting_duda"
    Jime = "setting_jime"
    Tron = "setting_tron"

    def get_yaml_file_path(self) -> path:
        path_player: path = os.path.join("data/settings/OneMatch", self.value + ".yaml")
        return path_player

    def get_match_settings_args(self) -> MatchSettingsArgs:
        match_args: MatchSettingsArgs = (
            parsley_coco.resolve_yaml_file_to_base_dataclass(
                yaml_path=self.get_yaml_file_path(), base_cls=MatchSettingsArgs
            )
        )
        return match_args



"""
This module contains the Script class which is responsible for launching scripts.
It handles computing execution time, profiling, and parsing arguments.
"""

import cProfile
import io
import os
import pprint
import pstats
import time
from pstats import SortKey
from typing import Any, Protocol, runtime_checkable

from parsley_coco import Parsley

from chipiron.scripts.script_args import BaseScriptArgs
from chipiron.utils import path
from chipiron.utils.dataclass import IsDataclass
from chipiron.utils.logger import chipiron_logger, suppress_logging
from chipiron.utils.small_tools import mkdir_if_not_existing


@runtime_checkable
class HasBaseScriptArgs(Protocol):
    """
    Protocol of generic ScriptArgs that contains the BaseScriptArgs
    """

    base_script_args: BaseScriptArgs


class Script[T_Dataclass: IsDataclass]:
    """
    The core Script class to launch scripts.
    Takes care of computing execution time, profiling, and parsing arguments.
    """

    start_time: float
    parser: Parsley[T_Dataclass]
    gui_args: dict[str, Any] | None
    profile: cProfile.Profile | None
    experiment_script_type_output_folder: path | None = None
    base_experiment_output_folder: path = "chipiron/scripts/"
    default_experiment_output_folder: path = "chipiron/scripts/default_output_folder"
    config_file_name: str | None

    def __init__(
        self,
        parser: Parsley[T_Dataclass],
        extra_args: dict[str, Any] | None = None,
        config_file_name: str | None = None,
    ) -> None:
        """
        Initializes the Script object.
        Starts the clock, the profiling, and parses arguments.

        Args:
            parser: An instance of MyParser used for parsing arguments.
            extra_args: Additional arguments to be passed to the parser.
        """
        self.start_time = time.time()  # start the clock
        self.parser = parser
        self.experiment_script_type_output_folder = None
        self.config_file_name = config_file_name
        self.extra_args = extra_args

        self.profile = None
        self.args: IsDataclass | None = None

    def initiate(
        self,
        experiment_output_folder: str | None = None,
    ) -> T_Dataclass:
        """
        Initiates the script by parsing arguments and converting them into a standardized dataclass.

        Args:
            args_dataclass_name: The type of the dataclass to convert the arguments into.
            experiment_output_folder: The base folder for experiment output. If None, uses the default value.

        Returns:
            The converted arguments as a dataclass.
        """

        # checking that the dataclass that will contain the script args contains BaseScriptArgs
        # assert issubclass(args_dataclass_name, HasBaseScriptArgs)

        if experiment_output_folder is not None:
            self.experiment_script_type_output_folder = experiment_output_folder
        else:
            self.experiment_script_type_output_folder = (
                self.default_experiment_output_folder
            )

        final_args: T_Dataclass = self.parser.parse_arguments(
            extra_args=self.extra_args, config_file_path=self.config_file_name
        )

        assert hasattr(final_args, "base_script_args")

        final_args.base_script_args.experiment_output_folder = os.path.join(
            self.experiment_script_type_output_folder,
            final_args.base_script_args.relative_script_instance_experiment_output_folder,
        )
        mkdir_if_not_existing(final_args.base_script_args.experiment_output_folder)
        mkdir_if_not_existing(
            os.path.join(
                final_args.base_script_args.experiment_output_folder,
                "inputs_and_parsing",
            )
        )

        self.parser.log_parser_info(
            final_args.base_script_args.experiment_output_folder
        )

        # activate profiling is if needed
        if final_args.base_script_args.profiling:
            self.profile = cProfile.Profile()
            self.profile.enable()

        self.args = final_args
        chipiron_logger.info(
            f"The args of the script are:\n{pprint.pformat(self.args)}"
        )

        return final_args

    def terminate(self) -> None:
        """
        Finishes the script by printing execution time and profiling information (if enabled).
        """
        chipiron_logger.info("terminate")
        if self.profile is not None:
            chipiron_logger.info(f"--- {time.time() - self.start_time} seconds ---")
            self.profile.disable()
            string_io = io.StringIO()
            sort_by = SortKey.CUMULATIVE
            stats = pstats.Stats(self.profile, stream=string_io).sort_stats(sort_by)
            stats.print_stats()
            chipiron_logger.info(string_io.getvalue())
            assert self.args is not None
            assert hasattr(self.args, "base_script_args")

            path_to_profiling_stats: path = os.path.join(
                self.args.base_script_args.experiment_output_folder,
                "profiling_output.stats",
            )
            path_to_profiling_stats_png: path = os.path.join(
                self.args.base_script_args.experiment_output_folder,
                "profiling_output.png",
            )
            stats.dump_stats(path_to_profiling_stats)
            os.popen(
                f"gprof2dot -f pstats {path_to_profiling_stats} | dot -Tpng -o {path_to_profiling_stats_png}"
            )

            end_time = time.time()

            chipiron_logger.info("The args of the script were:\n")
            pprint.pprint(self.args)
            chipiron_logger.info("Execution time", end_time - self.start_time)

    def run(self) -> None:
        """
        Runs the script.
        """



import typing
from dataclasses import asdict
from typing import Any, Iterator, Protocol, Self

import chess
import shakmaty_python_binding
import yaml

from chipiron.environments.chess.board.board_modification import BoardModificationP
from chipiron.environments.chess.move import moveUci
from chipiron.environments.chess.move.imove import moveKey

from .utils import FenPlusHistory, FenPlusMoveHistory, fen

# identifier that should be unique to any position
boardKey = tuple[
    int, int, int, int, int, int, bool, int, int | None, int, int, int, int, int
]

# identifier that removes the info (such as rounds) to count easily repeating position at difference round of the game
boardKeyWithoutCounters = tuple[
    int, int, int, int, int, int, bool, int, int | None, int, int, int
]


class LegalMoveKeyGeneratorP(Protocol):
    generated_moves: (
        dict[moveKey, chess.Move] | list[shakmaty_python_binding.MyMove] | None
    )
    all_generated_keys: list[moveKey] | None
    # whether to sort the legal_moves by their respective uci for easy comparison of various implementations
    sort_legal_moves: bool = False

    def __iter__(self) -> Iterator[moveKey]: ...

    def __next__(self) -> moveKey: ...

    def more_than_one_move(self) -> bool: ...

    def get_all(self) -> list[moveKey]: ...

    @property
    def fen(self) -> fen: ...


def compute_key(
    pawns: int,
    knights: int,
    bishops: int,
    rooks: int,
    queens: int,
    kings: int,
    turn: bool,
    castling_rights: int,
    ep_square: int | None,
    white: int,
    black: int,
    promoted: int,
    fullmove_number: int,
    halfmove_clock: int,
) -> boardKey:
    """
    Computes and returns a unique key representing the current state of the chess board.

    The key is computed by concatenating various attributes of the board, including the positions of pawns, knights,
    bishops, rooks, queens, and kings, as well as the current turn, castling rights, en passant square, halfmove clock,
    occupied squares for each color, promoted pieces, and the fullmove number.
    It is faster than calling the fen.
    Returns:
        str: A unique key representing the current state of the chess board.
    """
    string: boardKey = (
        pawns,
        knights,
        bishops,
        rooks,
        queens,
        kings,
        turn,
        castling_rights,
        ep_square,
        white,
        black,
        promoted,
        fullmove_number,
        halfmove_clock,
    )
    return string


# Note that we do not use Dict[Square, Piece] because of the rust version that would need to transform
# tuple[chess.PieceType, chess.Color] into Piece and would lose time
PieceMap = typing.Annotated[
    dict[chess.Square, tuple[chess.PieceType, chess.Color]],
    "a dictionary that list the pieces on the board",
]


class IBoard(Protocol):
    fast_representation_: boardKey
    legal_moves_: LegalMoveKeyGeneratorP

    def get_uci_from_move_key(self, move_key: moveKey) -> moveUci:
        assert self.legal_moves_.generated_moves is not None
        return self.legal_moves_.generated_moves[move_key].uci()

    def get_move_key_from_uci(self, move_uci: moveUci) -> moveKey:
        number_moves: int = len(self.legal_moves_.get_all())
        i: int
        assert self.legal_moves_.generated_moves is not None

        for i in range(number_moves):
            if self.legal_moves_.generated_moves[i].uci() == move_uci:
                return i

        raise Exception(
            "the code should not have reached this point: problem with"
            " legal moves / uci relation in boards object it seems"
        )

    def play_move_key(self, move: moveKey) -> BoardModificationP | None: ...

    def play_move_uci(self, move_uci: moveUci) -> BoardModificationP | None: ...

    @property
    def fen(self) -> fen: ...

    @property
    def move_history_stack(
        self,
    ) -> list[moveUci]: ...

    def ply(self) -> int:
        """
        Returns the number of half-moves (plies) that have been played on the board.

        :return: The number of half-moves played on the board.
        :rtype: int
        """
        ...

    @property
    def turn(self) -> chess.Color:
        """
        Get the current turn color.

        Returns:
            chess.Color: The color of the current turn.
        """
        ...

    def copy(self, stack: bool, deep_copy_legal_moves: bool = True) -> Self:
        """
        Create a copy of the current board.

        Args:
            stack (bool): Whether to copy the move stack as well.

        Returns:
            BoardChi: A new instance of the BoardChi class with the copied board.
        """
        ...

    def is_game_over(self) -> bool:
        """
        Check if the game is over.

        Returns:
            bool: True if the game is over, False otherwise.
        """
        ...

    @property
    def pawns(self) -> chess.Bitboard: ...

    @property
    def knights(self) -> chess.Bitboard: ...

    @property
    def bishops(self) -> chess.Bitboard: ...

    @property
    def rooks(self) -> chess.Bitboard: ...

    @property
    def queens(self) -> chess.Bitboard: ...

    @property
    def kings(self) -> chess.Bitboard: ...

    @property
    def white(self) -> chess.Bitboard: ...

    @property
    def black(self) -> chess.Bitboard: ...

    @property
    def halfmove_clock(self) -> int: ...

    @property
    def promoted(self) -> chess.Bitboard: ...

    @property
    def fullmove_number(self) -> int: ...

    @property
    def castling_rights(self) -> chess.Bitboard: ...

    @property
    def occupied(self) -> chess.Bitboard: ...

    def occupied_color(self, color: chess.Color) -> chess.Bitboard: ...

    def result(self, claim_draw: bool = False) -> str: ...

    def termination(self) -> chess.Termination | None: ...

    def dump(self, file: Any) -> None:
        # create minimal info for reconstruction that is the class FenPlusMoveHistory

        current_fen: fen = self.fen
        fen_plus_moves: FenPlusMoveHistory = FenPlusMoveHistory(
            current_fen=current_fen, historical_moves=self.move_history_stack
        )

        yaml.dump(asdict(fen_plus_moves), file, default_flow_style=False)

    @property
    def ep_square(self) -> int | None: ...

    @property
    def fast_representation(self) -> boardKey:
        """
        Returns a fast representation of the board.

        This method computes and returns a string representation of the board
        that can be quickly generated and used for various purposes.

        :return: A string representation of the board.
        :rtype: str
        """
        return self.fast_representation_

    @property
    def fast_representation_without_counters(self) -> boardKeyWithoutCounters:
        """
        Returns a fast representation of the board.

        This method computes and returns a string representation of the board
        that can be quickly generated and used for various purposes.

        :return: A string representation of the board.
        :rtype: str
        """
        assert self.fast_representation_ is not None
        return self.fast_representation_[:-2]

    def is_zeroing(self, move: moveKey) -> bool: ...

    def is_attacked(self, a_color: chess.Color) -> bool: ...

    @property
    def legal_moves(self) -> LegalMoveKeyGeneratorP: ...

    def number_of_pieces_on_the_board(self) -> int: ...

    def piece_map(self) -> dict[chess.Square, tuple[chess.PieceType, chess.Color]]: ...

    def has_kingside_castling_rights(self, color: chess.Color) -> bool: ...

    def has_queenside_castling_rights(self, color: chess.Color) -> bool: ...

    def print_chess_board(self) -> str: ...

    def tell_result(self) -> None: ...

    def into_fen_plus_history(self) -> FenPlusHistory: ...



"""
Module for the GameManagerFactory class.
"""

import queue
from dataclasses import dataclass, field
from typing import Any

import chess

import chipiron as ch
import chipiron.environments.chess.board as boards
import chipiron.players as players_m
from chipiron.games.game.game_args import GameArgs
from chipiron.players import PlayerFactoryArgs
from chipiron.players.boardevaluators.board_evaluator import (
    IGameBoardEvaluator,
    ObservableBoardEvaluator,
)
from chipiron.players.factory_higher_level import (
    MoveFunction,
    PlayerObserverFactory,
    create_player_observer_factory,
)
from chipiron.utils import path, seed
from chipiron.utils.communication.gui_player_message import PlayersColorToPlayerMessage
from chipiron.utils.dataclass import IsDataclass

from ...environments.chess.board.utils import FenPlusHistory
from ...environments.chess.move_factory import MoveFactory
from ...players.boardevaluators.table_base import SyzygyTable
from ...players.player_ids import PlayerConfigTag
from ...scripts.chipiron_args import ImplementationArgs
from .game import Game, ObservableGame
from .game_manager import GameManager
from .progress_collector import PlayerProgressCollectorObservable


@dataclass
class GameManagerFactory:
    """
    The GameManagerFactory creates GameManager once the players and rules have been decided.
    Calling create ask for the creation of a GameManager depending on args and players.
    This class is supposed to be independent of Match-related classes (contrarily to the GameArgsFactory)

    Args:
    syzygy_table (SyzygyTable | None): The syzygy table used for endgame tablebase lookups.
    game_manager_board_evaluator (IGameBoardEvaluator): The game board evaluator used for evaluating game positions.
    output_folder_path (path | None): The path to the output folder where game data will be saved.
    main_thread_mailbox (queue.Queue[IsDataclass]): The mailbox used for communication between processes.

    """

    # todo we might want to plit this into various part, like maybe a player factory, not sure, think about it

    syzygy_table: SyzygyTable[Any] | None
    output_folder_path: path | None
    main_thread_mailbox: queue.Queue[IsDataclass]
    game_manager_board_evaluator: IGameBoardEvaluator
    board_factory: boards.BoardFactory
    move_factory: MoveFactory
    implementation_args: ImplementationArgs
    universal_behavior: bool
    subscribers: list[queue.Queue[IsDataclass]] = field(default_factory=list)

    def create(
        self,
        args_game_manager: GameArgs,
        player_color_to_factory_args: dict[chess.Color, PlayerFactoryArgs],
        game_seed: seed,
    ) -> GameManager:
        """
        Create a GameManager with the given arguments

        Args:
            args_game_manager (GameArgs): the arguments of the game manager
            player_color_to_factory_args (dict[chess.Color, PlayerFactoryArgs]): the arguments of the players
            game_seed (int): the seed of the game

        Returns:
            the created GameManager
        """
        # useful if the logic of game generation gets complex
        # in the future, we might want the implementation detail to actually be modified during the
        # match in that case they would come arg_game_manager

        # CREATING THE BOARD
        starting_fen: str = args_game_manager.starting_position.get_fen()
        board: boards.IBoard = self.board_factory(
            fen_with_history=FenPlusHistory(current_fen=starting_fen)
        )
        if self.subscribers:
            for subscriber in self.subscribers:
                player_id_message: PlayersColorToPlayerMessage = (
                    PlayersColorToPlayerMessage(
                        player_color_to_factory_args=player_color_to_factory_args
                    )
                )

                subscriber.put(player_id_message)

        while not self.main_thread_mailbox.empty():
            self.main_thread_mailbox.get()

        # creating the game playing status
        game_playing_status: ch.games.GamePlayingStatus = ch.games.GamePlayingStatus()

        game: Game = Game(
            playing_status=game_playing_status, board=board, seed_=game_seed
        )
        observable_game: ObservableGame = ObservableGame(game=game)

        if self.subscribers:
            for subscriber in self.subscribers:
                observable_game.register_display(subscriber)

        # CREATING THE PLAYERS
        player_observer_factory: PlayerObserverFactory = create_player_observer_factory(
            each_player_has_its_own_thread=args_game_manager.each_player_has_its_own_thread,
            implementation_args=self.implementation_args,
            syzygy_table=self.syzygy_table,
            universal_behavior=self.universal_behavior,
        )

        player_progress_collector: PlayerProgressCollectorObservable = (
            PlayerProgressCollectorObservable(subscribers=self.subscribers)
        )

        players: list[players_m.GamePlayer | players_m.PlayerProcess] = []
        # Creating the players
        for player_color in chess.COLORS:
            player_factory_args: players_m.PlayerFactoryArgs = (
                player_color_to_factory_args[player_color]
            )

            # Human playing with gui does not need a player, as the playing moves will be generated directly
            # by the GUI and sent directly to the game_manager
            if player_factory_args.player_args.name != PlayerConfigTag.GUI_HUMAN:
                generic_player: players_m.GamePlayer | players_m.PlayerProcess
                move_function: MoveFunction
                generic_player, move_function = player_observer_factory(
                    player_color=player_color,
                    player_factory_args=player_factory_args,
                    main_thread_mailbox=self.main_thread_mailbox,
                    # player_progress_collector=player_progress_collector
                )
                players.append(generic_player)

                # registering to the observable board to get notification when it changes
                observable_game.register_player(move_function=move_function)

        player_color_to_id: dict[chess.Color, str] = {
            color: player_factory_args.player_args.name
            for color, player_factory_args in player_color_to_factory_args.items()
        }

        game_manager: GameManager
        game_manager = GameManager(
            game=observable_game,
            syzygy=self.syzygy_table,
            display_board_evaluator=self.game_manager_board_evaluator,
            output_folder_path=self.output_folder_path,
            args=args_game_manager,
            player_color_to_id=player_color_to_id,
            main_thread_mailbox=self.main_thread_mailbox,
            players=players,
            move_factory=self.move_factory,
            progress_collector=player_progress_collector,
        )

        return game_manager

    def subscribe(self, subscriber: queue.Queue[IsDataclass]) -> None:
        """
        Subscribe to the GameManagerFactory to get the PlayersColorToPlayerMessage
        As well as subscribing to the game_manager_board_evaluator to get the EvaluationMessage

        Args:
            subscriber: the subscriber queue
        """
        self.subscribers.append(subscriber)
        assert isinstance(self.game_manager_board_evaluator, ObservableBoardEvaluator)
        self.game_manager_board_evaluator.subscribe(subscriber)



"""
Module to check if an object is a dataclass
"""

from collections.abc import Iterable
from enum import Enum
from typing import Any, ClassVar, Protocol


class DataClass(Protocol):
    """
    Protocol to represent a dataclass.

    This protocol is used to check if an object is a dataclass by checking
    for the presence of the `__dataclass_fields__` attribute.
    """

    __dataclass_fields__: dict[str, Any]


class IsDataclass(Protocol):
    """
    Protocol to represent a dataclass.

    This protocol is used to check if an object is a dataclass by checking
    for the presence of the `__dataclass_fields__` attribute.
    """

    __dataclass_fields__: ClassVar[dict[Any, Any]]


# used as factory function in one of the option of as_dict in order to make the enum be replaced by str (useful for yamlization)
def custom_asdict_factory(data: Iterable[Any]) -> dict[Any, Any]:
    def convert_value(obj: Any) -> Any:
        if isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, list):
            return [convert_value(item) for item in obj]
        elif isinstance(obj, dict):
            return {k: convert_value(v) for k, v in obj.items()}
        else:
            return obj

    return dict((k, convert_value(v)) for k, v in data)



"""
Module to fetch, modify and convert the match settings and game settings.
"""

import os
from dataclasses import asdict, dataclass
from shutil import copyfile

import parsley_coco
import yaml

import chipiron as ch
import chipiron.games.game as game
from chipiron.games.match.match_args import MatchArgs
from chipiron.games.match.MatchTag import MatchConfigTag
from chipiron.utils import path

from .match_settings_args import MatchSettingsArgs


def fetch_match_games_args_convert_and_save(
    match_args: MatchArgs,
    profiling: bool = False,
    testing: bool = False,
    experiment_output_folder: path | None = None,
) -> tuple[MatchSettingsArgs, game.GameArgs]:
    """
    Fetches, modifies, and converts the match settings and game settings.

    Args:
        file_name_match_setting (path): The path to the match settings file.
        profiling (bool, optional): Flag indicating whether profiling is enabled. Defaults to False.
        testing (bool, optional): Flag indicating whether testing is enabled. Defaults to False.
        experiment_output_folder (path | None, optional): The path to the experiment output folder. Defaults to None.
        modification (dict[Any, Any] | None, optional): Dictionary of modifications to apply. Defaults to None.

    Returns:
        tuple[MatchSettingsArgs, game.GameArgs]: A tuple containing the match settings and game settings.
    """

    if profiling:
        path_match_setting: path = MatchConfigTag.Cubo.get_yaml_file_path()
        match_setting: MatchSettingsArgs = (
            parsley_coco.resolve_yaml_file_to_base_dataclass(
                yaml_path=path_match_setting,
                base_cls=MatchSettingsArgs,
            )
        )
        match_args.match_setting = match_setting

    assert isinstance(match_args.match_setting, MatchSettingsArgs)
    file_game: path = match_args.match_setting.game_setting_file
    path_game_setting: path = os.path.join("data/settings/GameSettings", file_game)

    if experiment_output_folder is not None:
        path_games: path = os.path.join(experiment_output_folder, "games")
        ch.tool.mkdir_if_not_existing(path_games)
        copyfile(
            src=path_game_setting, dst=os.path.join(experiment_output_folder, file_game)
        )
        with open(os.path.join(experiment_output_folder, "match_setting"), "w") as f:
            yaml.dump(asdict(match_args), f)

    game_file_path: path = os.path.join(
        "data/settings/GameSettings", match_args.match_setting.game_setting_file
    )
    args_game: game.GameArgs = parsley_coco.resolve_yaml_file_to_base_dataclass(
        yaml_path=game_file_path, base_cls=game.GameArgs, raise_error_with_nones=False
    )

    if profiling:
        args_game.max_half_moves = 1
    else:
        args_game.max_half_moves = None

    return match_args.match_setting, args_game



"""
Module defining the starting position arguments for the chess board.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Literal, Protocol

from chipiron.environments.chess.board.board_tools import convert_to_fen


class StartingPositionArgsType(str, Enum):
    """
    Enum class representing the type of starting position arguments.
    """

    fromFile = "from_file"
    fen = "fen"


@dataclass
class StartingPositionArgs(Protocol):
    """
    Dataclass representing the base class for starting position arguments.
    """

    type: StartingPositionArgsType

    def get_fen(self) -> str: ...


@dataclass
class FenStartingPositionArgs:
    """
    Dataclass representing the starting position arguments specified by FEN.
    """

    type: Literal[StartingPositionArgsType.fen]
    fen: str

    def get_fen(self) -> str:
        return self.fen


@dataclass
class FileStartingPositionArgs(StartingPositionArgs):
    """
    Dataclass representing the starting position arguments specified by a file.
    """

    type: Literal[StartingPositionArgsType.fromFile]
    file_name: str

    def get_fen(self) -> str:
        with open("data/starting_boards/" + self.file_name, "r") as f:
            ascii_board: str = str(f.read())
            fen: str = convert_to_fen(ascii_board)
        return fen


AllStartingPositionArgs = FenStartingPositionArgs | FileStartingPositionArgs



"""
This module defines the ValueUpdateInstructionsBlock class and a helper function to create instances of it.

The ValueUpdateInstructionsBlock class represents a block of update instructions for a tree value node in
 a move selector algorithm. It contains sets of moves that have been updated with new values,
  best moves, or are newly over.

The create_value_update_instructions_block function is a helper function that creates an instance of
 the ValueUpdateInstructionsBlock class with the specified update instructions.

"""

from dataclasses import dataclass, field
from typing import Self

from chipiron.environments.chess.move.imove import moveKey
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.algorithm_node import (
    AlgorithmNode,
)


@dataclass(slots=True)
class ValueUpdateInstructionsFromOneNode:
    node_sending_update: AlgorithmNode
    is_node_newly_over: bool
    new_value_for_node: bool
    new_best_move_for_node: bool


@dataclass(slots=True)
class ValueUpdateInstructionsTowardsOneParentNode:
    """
    Represents a block of value-update instructions intended to a specific node in the algorithm tree.

    Attributes:
        moves_with_updated_over (Set[AlgorithmNode]): Set of moves with updated 'over' value.
        moves_with_updated_value (Set[AlgorithmNode]): Set of moves with updated 'value' value.
        moves_with_updated_best_move (Set[AlgorithmNode]): Set of moves with updated 'best_move' value.
    """

    moves_with_updated_over: set[moveKey] = field(default_factory=set)
    moves_with_updated_value: set[moveKey] = field(default_factory=set)
    moves_with_updated_best_move: set[moveKey] = field(default_factory=set)

    def add_update_from_one_child_node(
        self,
        update_from_one_child_node: ValueUpdateInstructionsFromOneNode,
        move_from_parent_to_child: moveKey,
    ) -> None:
        if update_from_one_child_node.is_node_newly_over:
            self.moves_with_updated_over.add(move_from_parent_to_child)
        if update_from_one_child_node.new_value_for_node:
            self.moves_with_updated_value.add(move_from_parent_to_child)
        if update_from_one_child_node.new_best_move_for_node:
            self.moves_with_updated_best_move.add(move_from_parent_to_child)

    def add_update_toward_one_parent_node(self, another_update: Self) -> None:
        self.moves_with_updated_value = (
            self.moves_with_updated_value | another_update.moves_with_updated_over
        )
        self.moves_with_updated_over = (
            self.moves_with_updated_over | another_update.moves_with_updated_over
        )
        self.moves_with_updated_best_move = (
            self.moves_with_updated_best_move
            | another_update.moves_with_updated_best_move
        )

    def print_info(self) -> None:
        """
        Print information about the update instructions block.

        Returns:
            None
        """
        print("upInstructions printing")
        print(len(self.moves_with_updated_value), "moves_with_updated_value", end=" ")
        for move in self.moves_with_updated_value:
            print(move, end=" ")
        print(
            "\n",
            len(self.moves_with_updated_best_move),
            "moves_with_updated_best_move:",
            end=" ",
        )
        for move in self.moves_with_updated_best_move:
            print(move, end=" ")
        print(
            "\n", len(self.moves_with_updated_over), "moves_with_updated_over", end=" "
        )
        for move in self.moves_with_updated_over:
            print(move, end=" ")
        print()

    def empty(self) -> bool:
        """
        Check if all the components of the update instructions block are empty.

        Returns:
            bool: True if all components are empty, False otherwise.
        """
        empty_bool = (
            not bool(self.moves_with_updated_value)
            and not bool(self.moves_with_updated_best_move)
            and not bool(self.moves_with_updated_over)
        )
        return empty_bool



from dataclasses import dataclass
from datetime import datetime

from chipiron.utils import path


@dataclass
class BaseScriptArgs:
    """
    Dataclass representing the arguments for the Script class.
    """

    # whether the script is profiling computation usage
    profiling: bool = False

    # whether the script is testing the code (using pytest for instance)
    testing: bool = False

    # whether the script should force the behavior of the algorithm be only dependent on the seed and number
    # and on no other parameter ( could happen that using some external functions/modules add some randomness)
    universal_behavior: bool = False

    # if experiment_output_folder is specified it is used as is. If not it is build from default path plus the
    # relative_script_instance_experiment_output_folder

    # the folder where to output the results
    experiment_output_folder: path | None = None

    # specific folder name  where to output the results (if none it is set to time and day in post init)
    relative_script_instance_experiment_output_folder: path | None = None

    # the seed
    seed: int = 0

    def __post_init__(self) -> None:

        # if relative_script_instance_experiment_output_folde is not set, it gets time and day
        if self.relative_script_instance_experiment_output_folder is None:
            now = datetime.now()  # current date and time
            self.relative_script_instance_experiment_output_folder = now.strftime(
                "%A-%m-%d-%Y--%H:%M:%S:%f"
            )



"""
This module defines the IScript interface, which serves as the interface for scripts in the application.
"""

from typing import Any, Protocol

from .script import Script


class IScript(Protocol):
    """
    The interface for scripts in the application.
    """

    def __init__(self, base_script: Script[Any]) -> None:
        """
        Initializes the IScript object.

        Args:
            base_script (Script): The base script object.

        Returns:
            None
        """

    def run(self) -> None:
        """
        Runs the script.

        Returns:
            None
        """

    def terminate(self) -> None:
        """
        Finishes the script. Performs profiling or timing.

        Returns:
            None
        """



"""
This module provides a factory function for creating node evaluators based on different types of board evaluators.
"""

import sys
from typing import Any, TypeAlias

import chipiron.players.boardevaluators.basic_evaluation as basic_evaluation
from chipiron.players.boardevaluators.neural_networks.factory import (
    create_nn_board_eval_from_nn_parameters_file_and_existing_model,
)
from chipiron.players.boardevaluators.neural_networks.nn_board_evaluator import (
    NNBoardEvaluator,
)
from chipiron.players.boardevaluators.table_base.syzygy_table import SyzygyTable

from . import neural_networks
from .all_node_evaluators import NodeEvaluatorTypes
from .neural_networks.nn_node_evaluator import NNNodeEvaluator
from .node_evaluator import NodeEvaluator
from .node_evaluator_args import BasicEvaluationNodeEvaluatorArgs, NodeEvaluatorArgs

AllNodeEvaluatorArgs: TypeAlias = (
    neural_networks.NeuralNetNodeEvalArgs | BasicEvaluationNodeEvaluatorArgs
)


def create_node_evaluator(
    arg_board_evaluator: AllNodeEvaluatorArgs, syzygy: SyzygyTable[Any] | None
) -> NodeEvaluator:
    """
    Create a node evaluator based on the given board evaluator argument and syzygy table.

    Args:
        arg_board_evaluator (AllNodeEvaluatorArgs): The argument for the board evaluator.
        syzygy (SyzygyTable | None): The syzygy table to be used for evaluation.

    Returns:
        NodeEvaluator: The created node evaluator.

    Raises:
        SystemExit: If the given board evaluator type is not found.

    """
    if arg_board_evaluator.syzygy_evaluation:
        syzygy_ = syzygy
    else:
        syzygy_ = None

    node_evaluator: NodeEvaluator

    match arg_board_evaluator.type:
        case NodeEvaluatorTypes.BasicEvaluation:
            board_evaluator: basic_evaluation.BasicEvaluation = (
                basic_evaluation.BasicEvaluation()
            )
            node_evaluator = NodeEvaluator(
                board_evaluator=board_evaluator, syzygy=syzygy_
            )
        case NodeEvaluatorTypes.NeuralNetwork:
            assert isinstance(
                arg_board_evaluator, neural_networks.NeuralNetNodeEvalArgs
            )
            board_evaluator_nn: NNBoardEvaluator
            board_evaluator_nn = create_nn_board_eval_from_nn_parameters_file_and_existing_model(
                model_weights_file_name=arg_board_evaluator.neural_nets_model_and_architecture.model_weights_file_name,
                nn_architecture_args=arg_board_evaluator.neural_nets_model_and_architecture.nn_architecture_args,
            )
            node_evaluator = NNNodeEvaluator(
                nn_board_evaluator=board_evaluator_nn, syzygy=syzygy_
            )
        case other:
            sys.exit(f"Node Board Eval: can not find {other} in file {__name__}")

    return node_evaluator



"""
This module contains functions for testing the indices used in the move selector tree.

The main functions in this module are:
- `make_tree_from_file`: Creates a move and value tree from a YAML file.
- `check_from_file`: Compares the indices computed from the tree with the expected indices from a YAML file.
- `check_index`: Tests the indices for a specific index computation type and tree file.
- `test_indices`: Runs the index tests for multiple index computation types and tree files.
"""

from enum import Enum
from math import isclose
from typing import Any

import chess
import yaml

import chipiron.players.move_selector.treevalue.node_factory as node_factory
import chipiron.players.move_selector.treevalue.search_factory as search_factories
import chipiron.players.move_selector.treevalue.tree_manager as tree_manager
import chipiron.players.move_selector.treevalue.trees as trees
from chipiron.environments.chess.board.factory import (
    create_board_chi_from_pychess_board,
)
from chipiron.players.move_selector.treevalue.indices.index_manager.factory import (
    create_exploration_index_manager,
)
from chipiron.players.move_selector.treevalue.indices.index_manager.node_exploration_manager import (
    NodeExplorationIndexManager,
    update_all_indices,
)
from chipiron.players.move_selector.treevalue.indices.node_indices.index_types import (
    IndexComputationType,
)
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.algorithm_node import (
    AlgorithmNode,
)
from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode
from chipiron.players.move_selector.treevalue.tree_manager.tree_expander import (
    TreeExpansion,
    TreeExpansions,
)
from chipiron.players.move_selector.treevalue.trees.descendants import RangedDescendants
from chipiron.players.move_selector.treevalue.trees.move_and_value_tree import (
    MoveAndValueTree,
)
from chipiron.utils.small_tools import path


class TestResult(Enum):
    """
    Enumeration for the test results.
    """

    __test__ = False
    PASSED = 0
    FAILED = 1
    WARNING = 2


def make_tree_from_file(
    file_path: path, index_computation: IndexComputationType
) -> MoveAndValueTree:
    """
    Creates a move and value tree from a file.

    Args:
        file_path (path): The path to the file containing the tree data.
        index_computation (IndexComputationType): The type of index computation to use.

    Returns:
        MoveAndValueTree: The created move and value tree.
    """

    # atm it is very ad hoc to test index so takes a lots of shortcut, will be made more general when needed
    with open(file_path, "r") as file:
        tree_yaml = yaml.safe_load(file)
    print("tree", tree_yaml)
    yaml_nodes = tree_yaml["nodes"]

    node_factory_name: str = "Base_with_algorithm_tree_node"
    tree_node_factory: node_factory.Base[Any] = node_factory.create_node_factory(
        node_factory_name=node_factory_name
    )

    search_factory: search_factories.SearchFactoryP = search_factories.SearchFactory(
        node_selector_args=None,
        opening_type=None,
        random_generator=None,
        index_computation=index_computation,
    )

    algorithm_node_factory: node_factory.AlgorithmNodeFactory
    algorithm_node_factory = node_factory.AlgorithmNodeFactory(
        tree_node_factory=tree_node_factory,
        board_representation_factory=None,
        exploration_index_data_create=search_factory.node_index_create,
    )
    descendants: RangedDescendants = RangedDescendants()

    algo_tree_manager: tree_manager.AlgorithmNodeTreeManager = (
        tree_manager.create_algorithm_node_tree_manager(
            node_evaluator=None,
            algorithm_node_factory=algorithm_node_factory,
            index_computation=index_computation,
            index_updater=None,
        )
    )

    half_moves = {}
    id_nodes = {}
    for yaml_node in yaml_nodes:
        # print('yaml_node[id] ',yaml_node['id'] )
        if yaml_node["id"] == 0:
            tree_expansions = TreeExpansions()

            board = chess.Board.from_chess960_pos(yaml_node["id"])
            board.turn = chess.WHITE
            board_chi = create_board_chi_from_pychess_board(chess_board=board)

            root_node: ITreeNode[Any] = algorithm_node_factory.create(
                board=board_chi,
                half_move=0,
                count=yaml_node["id"],
                parent_node=None,
                move_from_parent=None,
                modifications=None,
            )
            assert isinstance(root_node, AlgorithmNode)
            root_node.minmax_evaluation.value_white_minmax = yaml_node["value"]
            half_moves[yaml_node["id"]] = 0
            id_nodes[yaml_node["id"]] = root_node

            descendants.add_descendant(root_node)
            move_and_value_tree: MoveAndValueTree = MoveAndValueTree(
                root_node=root_node, descendants=descendants
            )
            tree_expansions.add(
                TreeExpansion(
                    child_node=root_node,
                    parent_node=None,
                    board_modifications=None,
                    creation_child_node=True,
                    move=None,
                )
            )
            root_node.tree_node.all_legal_moves_generated = True
            # algo_tree_manager.update_backward(tree_expansions=tree_expansions)

        else:
            tree_expansions = TreeExpansions()
            first_parent = yaml_node["parents"]
            half_move = half_moves[first_parent] + 1
            half_moves[yaml_node["id"]] = half_move
            parent_node = id_nodes[first_parent]
            board = chess.Board.from_chess960_pos(yaml_node["id"])
            board.turn = not parent_node.tree_node.board_.turn
            board_chi = create_board_chi_from_pychess_board(chess_board=board)

            tree_expansion: TreeExpansion = algo_tree_manager.tree_manager.open_node(
                tree=move_and_value_tree,
                parent_node=parent_node,
                board=board_chi,
                modifications=None,
                move=yaml_node["id"],
            )
            tree_expansions.add(
                TreeExpansion(
                    child_node=tree_expansion.child_node,
                    parent_node=tree_expansion.parent_node,
                    board_modifications=tree_expansion.board_modifications,
                    creation_child_node=tree_expansion.creation_child_node,
                    move=yaml_node["id"],
                )
            )
            assert isinstance(tree_expansion.child_node, AlgorithmNode)
            tree_expansion.child_node.tree_node.all_legal_moves_generated = True
            id_nodes[yaml_node["id"]] = tree_expansion.child_node
            tree_expansion.child_node.minmax_evaluation.value_white_minmax = yaml_node[
                "value"
            ]
            tree_expansion.child_node.minmax_evaluation.value_white_evaluator = (
                yaml_node["value"]
            )
            assert tree_expansion.move is not None
            parent_node.minmax_evaluation.moves_not_over.append(tree_expansion.move)
            algo_tree_manager.update_backward(tree_expansions=tree_expansions)

    # print('move_and_value_tree', move_and_value_tree.descendants)
    return move_and_value_tree


def check_from_file(file_path: path, tree: MoveAndValueTree) -> None:
    """
    Check the values in the given file against the values in the tree.

    Args:
        file_path (str): The path to the file containing the values to check.
        tree (MoveAndValueTree): The tree containing the values to compare against.

    Returns:
        None
    """
    with open(file_path, "r") as file:
        tree_yaml = yaml.safe_load(file)
    print("tree", tree_yaml)
    yaml_nodes = tree_yaml["nodes"]

    tree_nodes: trees.RangedDescendants = tree.descendants

    half_move: int
    for half_move in tree_nodes:
        # todo how are we sure that the hm comes in order?
        # print('hmv', half_move)
        parent_node: ITreeNode[Any]
        for parent_node in tree_nodes[half_move].values():
            assert isinstance(parent_node, AlgorithmNode)
            yaml_index = eval(str(yaml_nodes[parent_node.id]["index"]))
            assert parent_node.exploration_index_data is not None
            print(
                f"id {parent_node.id} expected value {yaml_index} "
                f"|| computed value {parent_node.exploration_index_data.index}"
                f' {type(yaml_nodes[parent_node.id]["index"])}'
                f" {type(parent_node.exploration_index_data.index)}"
                f"{(yaml_index == parent_node.exploration_index_data.index)}"
            )
            if yaml_index is None:
                assert parent_node.exploration_index_data.index is None
            else:
                assert parent_node.exploration_index_data.index is not None
                assert isclose(
                    yaml_index, parent_node.exploration_index_data.index, abs_tol=1e-8
                )


def check_index(index_computation: IndexComputationType, tree_file: path) -> TestResult:
    """
    Checks the index for a given tree file and index computation type.

    Args:
        index_computation (IndexComputationType): The type of index computation.
        tree_file (path): The path to the tree file.

    Returns:
        TestResult: The result of the index check.

    Raises:
        None

    """
    try:
        tree_path = f"data/trees/{tree_file}/{tree_file}_{index_computation.value}.yaml"
    except Exception:
        print(f"!!!!!!Warning!!!!! : no testing file {tree_path}")
        return TestResult.WARNING

    tree_path = f"data/trees/{tree_file}/{tree_file}.yaml"
    tree: MoveAndValueTree = make_tree_from_file(
        index_computation=index_computation, file_path=tree_path
    )

    index_manager: NodeExplorationIndexManager = create_exploration_index_manager(
        index_computation=index_computation
    )

    print("index_manager", index_manager)
    update_all_indices(tree, index_manager)
    # print_all_indices(
    #    tree
    # )
    file_index = f"data/trees/{tree_file}/{tree_file}_{index_computation.value}.yaml"
    check_from_file(file_path=file_index, tree=tree)

    return TestResult.PASSED


def test_indices() -> None:
    """
    Test the index computations on multiple tree files.

    This function iterates over a list of index computations and tree files,
    and performs a test for each combination. The results of the tests are
    stored in a dictionary.

    Returns:
        None
    """
    index_computations: list[IndexComputationType] = [
        IndexComputationType.MinGlobalChange,
        IndexComputationType.RecurZipf,
        IndexComputationType.MinLocalChange,
    ]

    tree_files = ["tree_1", "tree_2"]

    results: dict[TestResult, int] = {}
    for tree_file in tree_files:

        if tree_file == "tree_2":
            index_computations_ = [
                IndexComputationType.MinGlobalChange,
                IndexComputationType.MinLocalChange,
            ]
        else:
            index_computations_ = index_computations

        for index_computation in index_computations_:
            print(f"---testing {index_computation} on {tree_file}")
            res: TestResult = check_index(
                index_computation=index_computation, tree_file=tree_file
            )
            if res in results:
                results[res] += 1
            else:
                results[res] = 1
    print(f"finished Test: {results}")
    assert results[TestResult.PASSED] == 5


if __name__ == "__main__":
    test_indices()



# logger_module.py

import logging
from contextlib import contextmanager
from typing import Generator

chipiron_logger: logging.Logger = logging.getLogger("chipiron_app")
chipiron_logger.setLevel(logging.DEBUG)

# Avoid duplicate handlers if this module is imported multiple times
if not chipiron_logger.handlers:
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)

    formatter = logging.Formatter("[%(levelname)s] %(message)s")
    console_handler.setFormatter(formatter)

    chipiron_logger.addHandler(console_handler)
    chipiron_logger.propagate = False


@contextmanager
def suppress_logging(
    logger: logging.Logger, level: int = logging.WARNING
) -> Generator[None, None, None]:
    previous_level = logger.level
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(previous_level)



from typing import Any

from chipiron.players.player_ids import PlayerConfigTag
from chipiron.scripts.factory import create_script
from chipiron.scripts.script_gui.script_gui_custom import generate_inputs

output_list: list[dict[str, Any]] = [
    {
        "type": "play_against_chipiron",
        "strength": 1,
        "color_human": "White",
        "chipi_algo": PlayerConfigTag.UNIFORM,
    },
    {
        "type": "play_against_chipiron",
        "strength": 2,
        "color_human": "Black",
        "chipi_algo": PlayerConfigTag.CHIPIRON,
    },
    {
        "type": "play_against_chipiron",
        "strength": 2,
        "color_human": "Black",
        "chipi_algo": PlayerConfigTag.SEQUOOL,
    },
    {"type": "play_two_humans"},
    {"type": "watch_a_game"},
]


def test_script_gui() -> None:
    for output in output_list:
        script_type, gui_extra_args, config_file_name = generate_inputs(output=output)

        # creating the script object from its name and arguments
        create_script(
            script_type=script_type,
            extra_args=gui_extra_args,
            config_file_name=config_file_name,
            should_parse_command_line_arguments=False,
        )


if __name__ == "__main__":
    test_script_gui()
    print("all test passed!")



"""
This module provides classes and functions for neural network node evaluation in the tree value move selector.

Classes:
- NeuralNetNodeEvalArgs: A class representing the arguments for neural network node evaluation.

"""

from .factory import NeuralNetNodeEvalArgs

__all__ = ["NeuralNetNodeEvalArgs"]



"""
This module contains the implementation of the StockfishPlayer class, which is a move selector that uses the Stockfish chess engine to recommend moves.

The StockfishPlayer class is a dataclass that represents a player that selects moves using the Stockfish engine. It has the following attributes:
- type: A literal value representing the type of move selector (in this case, MoveSelectorTypes.Stockfish).
- depth: An integer representing the depth to which Stockfish should search for moves (default is 20).
- time_limit: A float representing the time limit (in seconds) for Stockfish to search for moves (default is 0.1).
- engine: An instance of the Stockfish engine.

The StockfishPlayer class has the following methods:
- select_move: Selects a move based on the given board state and move seed. Returns a MoveRecommendation object.
- print_info: Prints the type of move selector.

Note: The Stockfish engine is initialized lazily when the first move is selected, and it is automatically closed after each move is selected.
"""

from dataclasses import dataclass
from typing import Any, Literal

import chess.engine

import chipiron.environments.chess.board as boards

from ...environments.chess import BoardChi
from ...environments.chess.board import create_board_chi
from ...environments.chess.board.utils import FenPlusHistory
from .move_selector import MoveRecommendation
from .move_selector_types import MoveSelectorTypes


@dataclass
class StockfishPlayer:
    """
    A player that selects moves using the Stockfish chess engine.

    Attributes:
        type (Literal[MoveSelectorTypes.Stockfish]): The type of move selector (for serialization).
        depth (int): The depth to which Stockfish should search for moves.
        time_limit (float): The time limit (in seconds) for Stockfish to search for moves.
        engine (Any): The Stockfish chess engine instance.

    Methods:
        select_move(board: boards.BoardChi, move_seed: int) -> MoveRecommendation:
            Selects a move based on the given board state and move seed.

        print_info() -> None:
            Prints the type of move selector.
    """

    type: Literal[MoveSelectorTypes.Stockfish]  # for serialization
    depth: int = 20
    time_limit: float = 0.1
    engine: Any = None

    def select_move(self, board: boards.IBoard, move_seed: int) -> MoveRecommendation:
        """
        Selects a move based on the given board state and move seed.

        Args:
            board (boards.BoardChi): The current board state.
            move_seed (int): The seed for move selection.

        Returns:
            MoveRecommendation: A MoveRecommendation object representing the selected move.
        """

        if self.engine is None:
            # if this object is created in the init then sending the object
            # self.engine = chess.engine.SimpleEngine.popen_uci(
            #    # TODO: should we remove the hardcoding
            #    r"stockfish/stockfish_14.1_linux_x64/stockfish_14.1_linux_x64")
            self.engine = chess.engine.SimpleEngine.popen_uci(
                # TODO: should we remove the hardcoding
                r"stockfish/stockfish/stockfish-ubuntu-x86-64-avx2"
            )

        # transform the board
        board_chi: BoardChi = create_board_chi(
            fen_with_history=FenPlusHistory(
                current_fen=board.fen,
                historical_moves=board.move_history_stack,
                # note that we do not give here historical_boards, hope this does not create but related to 3 fold repetition computation
            )
        )
        result = self.engine.play(
            board_chi.chess_board, chess.engine.Limit(self.time_limit)
        )

        self.engine.quit()
        self.engine = None
        return MoveRecommendation(move=result.move.uci())

    def print_info(self) -> None:
        """
        Prints the type of move selector.
        """
        print(self.type)



from enum import Enum
from typing import Callable

import torch.nn as nn


class NNModelType(str, Enum):
    MultiLayerPerceptron = "multi_layer_perceptron"
    Transformer = "transformer"


class ActivationFunctionType(str, Enum):
    TangentHyperbolic = "hyperbolic_tangent"
    ParametricRelu = "parametric_relu"
    Relu = "relu"


activation_map: dict[ActivationFunctionType, Callable[[], nn.Module]] = {
    ActivationFunctionType.Relu: nn.ReLU,
    ActivationFunctionType.TangentHyperbolic: nn.Tanh,
    ActivationFunctionType.ParametricRelu: nn.PReLU,
}



"""
This module defines the different types of node evaluators used in the tree value calculation.
"""

from enum import Enum


class NodeEvaluatorTypes(str, Enum):
    """
    Enum class representing different types of node evaluators.
    """

    NeuralNetwork = "neural_network"
    BasicEvaluation = "basic_evaluation"



"""
This script defines a graphical user interface (GUI) for interacting with the chipiron program.
The GUI allows the user to select various options and perform different actions such as playing against the chipiron AI, watching a game, or visualizing a tree.

The script_gui function creates the GUI window and handles the user interactions.
The play_against_chipiron, watch_a_game, and visualize_a_tree functions are called when the corresponding buttons are clicked.

Note: This script requires the customtkinter and chipiron modules to be installed.
"""

# TODO switch to pygame
from typing import Any

import customtkinter as ctk
from parsley_coco import make_partial_dataclass_with_optional_paths

from chipiron import scripts
from chipiron.games.match.match_args import MatchArgs
from chipiron.games.match.MatchTag import MatchConfigTag
from chipiron.players import PlayerArgs
from chipiron.players.move_selector.move_selector_types import MoveSelectorTypes
from chipiron.players.move_selector.treevalue import TreeAndValuePlayerArgs
from chipiron.players.move_selector.treevalue.progress_monitor.progress_monitor import (
    StoppingCriterionTypes,
    TreeMoveLimitArgs,
)
from chipiron.players.player_ids import PlayerConfigTag
from chipiron.scripts.one_match.one_match import MatchScriptArgs
from chipiron.scripts.script_args import BaseScriptArgs


def script_gui() -> tuple[scripts.ScriptType, dict[str, Any], str]:
    """
    Creates a graphical user interface (GUI) for interacting with the chipiron program.

    Returns:
        A tuple containing the script type and the arguments for the selected action.
    """
    root = ctk.CTk()
    output: dict[str, Any] = {}
    # place a label on the root window
    root.title("chipiron")

    # frm = ctk.CTkFrame(root, padding=10)

    window_width: int = 800
    window_height: int = 300

    # get the screen dimension
    screen_width: int = root.winfo_screenwidth()
    screen_height: int = root.winfo_screenheight()

    # find the center point
    center_x: int = int(screen_width / 2 - window_width / 2)
    center_y: int = int(screen_height / 2 - window_height / 2)

    # set the position of the window to the center of the screen
    root.geometry(f"{window_width}x{window_height}+{center_x}+{center_y}")
    # root.iconbitmap('download.jpeg')

    message = ctk.CTkLabel(root, text="What to do?")
    message.grid(column=0, row=0)

    # exit button
    exit_button = ctk.CTkButton(root, text="Exit", command=lambda: root.quit())

    message = ctk.CTkLabel(root, text="Play ")
    message.grid(column=0, row=2)

    # Create the list of options
    color_options_list: list[str] = ["White", "Black"]

    # Variable to keep track of the option
    # selected in OptionMenu
    # color_choice_human = ctk.StringVar(root)
    color_choice_human = ctk.StringVar(value="White")  # set initial value

    # Set the default value of the variable
    # color_choice_human.set("White")

    # Create the option menu widget and passing
    # the options_list and value_inside to it.
    strength_menu = ctk.CTkOptionMenu(
        master=root, values=color_options_list, variable=color_choice_human
    )
    strength_menu.grid(column=1, row=2)

    message = ctk.CTkLabel(root, text=" against ")
    message.grid(column=2, row=2)

    # Create the list of options
    chipi_algo_options_list: list[PlayerConfigTag] = [
        PlayerConfigTag.RECUR_ZIPF_BASE_3,
        PlayerConfigTag.UNIFORM,
        PlayerConfigTag.SEQUOOL,
    ]

    # Variable to keep track of the option
    # selected in OptionMenu
    chipi_algo_choice = ctk.StringVar(
        value=PlayerConfigTag.RECUR_ZIPF_BASE_3
    )  # set initial value

    # chipi_algo_choice = tk.StringVar(root)

    # Set the default value of the variable
    # chipi_algo_choice.set("RecurZipfBase3")

    # Create the option menu widget and passing
    # the options_list and value_inside to it.
    strength_menu = ctk.CTkOptionMenu(
        master=root, values=chipi_algo_options_list, variable=chipi_algo_choice
    )
    strength_menu.grid(column=3, row=2)

    message = ctk.CTkLabel(root, text="  strength: ")

    message.grid(column=5, row=2, padx=10, pady=10)

    # Create the list of options
    options_list = ["1", "2", "3", "4", "5"]

    # Variable to keep track of the option
    # selected in OptionMenu
    strength_value = ctk.StringVar(value="1")  # set initial value

    # strength_value = tk.StringVar(root)

    # Set the default value of the variable
    # strength_value.set("1")

    # Create the option menu widget and passing
    # the options_list and value_inside to it.
    strength_menu = ctk.CTkOptionMenu(
        master=root, variable=strength_value, values=options_list
    )
    strength_menu.grid(column=5, row=2, padx=10, pady=10)

    # play button
    play_against_chipiron_button: ctk.CTkButton = ctk.CTkButton(
        root,
        text="!Play!",
        command=lambda: [
            play_against_chipiron(
                output,
                strength=strength_value,
                color=color_choice_human,
                chipi_algo=chipi_algo_choice,
            ),
            root.destroy(),
        ],
    )

    # play_two_humans button
    play_two_humans_button: ctk.CTkButton = ctk.CTkButton(
        root,
        text="Play between Two Humans",
        command=lambda: [play_two_humans(output), root.destroy()],
    )

    # watch button
    watch_a_game_button: ctk.CTkButton = ctk.CTkButton(
        root,
        text="Watch a game",
        command=lambda: [watch_a_game(output), root.destroy()],
    )

    # visualize button
    visualize_a_tree_button: ctk.CTkButton = ctk.CTkButton(
        root,
        text="Visualize a tree",
        command=lambda: [visualize_a_tree(output), root.destroy()],
    )

    play_against_chipiron_button.grid(row=2, column=6, padx=10, pady=10)
    play_two_humans_button.grid(row=4, column=0, padx=10, pady=10)
    watch_a_game_button.grid(row=6, column=0, padx=10, pady=10)
    visualize_a_tree_button.grid(row=8, column=0, padx=10, pady=10)
    exit_button.grid(row=10, column=0, padx=10, pady=10)

    root.mainloop()

    return generate_inputs(output=output)


def generate_inputs(
    output: dict[str, Any],
) -> tuple[scripts.ScriptType, dict[str, Any], str]:
    script_type: scripts.ScriptType

    PartialOpMatchScriptArgs = make_partial_dataclass_with_optional_paths(
        cls=MatchScriptArgs
    )
    PartialOpMatchArgs = make_partial_dataclass_with_optional_paths(cls=MatchArgs)
    PartialOpPlayerArgs = make_partial_dataclass_with_optional_paths(cls=PlayerArgs)

    PartialOpBaseScriptArgs = make_partial_dataclass_with_optional_paths(
        cls=BaseScriptArgs
    )

    PartialOpTreeAndValuePlayerArgs = make_partial_dataclass_with_optional_paths(
        cls=TreeAndValuePlayerArgs
    )
    PartialOpTreeMoveLimitArgs = make_partial_dataclass_with_optional_paths(
        cls=TreeMoveLimitArgs
    )

    config_file_name: str
    match output["type"]:
        case "play_against_chipiron":
            tree_move_limit = 4 * 10 ** output["strength"]
            gui_args = PartialOpMatchScriptArgs(
                gui=True,
                base_script_args=PartialOpBaseScriptArgs(profiling=False, seed=0),
                match_args=PartialOpMatchArgs(match_setting=MatchConfigTag.Duda),
            )
            config_file_name = "chipiron/scripts/one_match/inputs/human_play_against_computer/exp_options.yaml"

            if output["color_human"] == "White":
                gui_args.match_args.player_one = PlayerConfigTag.GUI_HUMAN
                gui_args.match_args.player_two = PlayerConfigTag(output["chipi_algo"])
                gui_args.match_args.player_two_overwrite = PartialOpPlayerArgs(
                    main_move_selector=PartialOpTreeAndValuePlayerArgs(
                        type=MoveSelectorTypes.TreeAndValue,
                        stopping_criterion=PartialOpTreeMoveLimitArgs(
                            type=StoppingCriterionTypes.TreeMoveLimit,
                            tree_move_limit=tree_move_limit,
                        ),
                    )
                )
            else:
                gui_args.match_args.player_one = PlayerConfigTag(output["chipi_algo"])
                gui_args.match_args.player_two = PlayerConfigTag.GUI_HUMAN
                gui_args.match_args.player_one_overwrite = PartialOpPlayerArgs(
                    main_move_selector=PartialOpTreeAndValuePlayerArgs(
                        stopping_criterion=PartialOpTreeMoveLimitArgs(
                            type=StoppingCriterionTypes.TreeMoveLimit,
                            tree_move_limit=tree_move_limit,
                        )
                    )
                )
            script_type = scripts.ScriptType.OneMatch
        case "play_two_humans":
            config_file_name = "chipiron/scripts/one_match/inputs/human_play_against_human/exp_options.yaml"

            gui_args = PartialOpMatchScriptArgs(
                gui=True,
                base_script_args=PartialOpBaseScriptArgs(profiling=False, seed=0),
                match_args=PartialOpMatchArgs(match_setting=MatchConfigTag.Duda),
            )

            gui_args.match_args.player_one = PlayerConfigTag.GUI_HUMAN
            gui_args.match_args.player_two = PlayerConfigTag.GUI_HUMAN
            script_type = scripts.ScriptType.OneMatch
        case "watch_a_game":
            config_file_name = (
                "chipiron/scripts/one_match/inputs/watch_a_game/exp_options.yaml"
            )
            gui_args = PartialOpMatchScriptArgs(
                gui=True,
                base_script_args=PartialOpBaseScriptArgs(profiling=False, seed=0),
                match_args=PartialOpMatchArgs(
                    match_setting=MatchConfigTag.Duda,
                    player_one=PlayerConfigTag.RECUR_ZIPF_BASE_3,
                    player_two=PlayerConfigTag.RECUR_ZIPF_BASE_3,
                ),
            )

            script_type = scripts.ScriptType.OneMatch
        case "tree_visualization":
            config_file_name = "scripts/tree_visualization/inputs/base/exp_options.yaml"
            gui_args = None
            script_type = scripts.ScriptType.TreeVisualization
        case other:
            raise Exception(f"Not a good name: {other}")

    print(
        f"Gui choices: the script name is {script_type} and the args are {gui_args} and config file {config_file_name}"
    )
    return script_type, gui_args, config_file_name


def play_against_chipiron(
    output: dict[str, Any],
    strength: ctk.StringVar,
    color: ctk.StringVar,
    chipi_algo: ctk.StringVar,
) -> bool:
    """
    Callback function for the "Play" button.
    Sets the output dictionary with the selected options for playing against the chipiron AI.

    Args:
        output: The output dictionary to store the selected options.
        strength: The selected strength level.
        color: The selected color.
        chipi_algo: The selected chipiron algorithm.

    Returns:
        True.
    """
    output["type"] = "play_against_chipiron"
    output["strength"] = int(strength.get())
    output["color_human"] = str(color.get())
    output["chipi_algo"] = PlayerConfigTag(str(chipi_algo.get()))
    return True


def watch_a_game(output: dict[str, Any]) -> bool:
    """
    Callback function for the "Watch a game" button.
    Sets the output dictionary with the selected options for watching a game.

    Args:
        output: The output dictionary to store the selected options.

    Returns:
        True.
    """
    output["type"] = "watch_a_game"
    return True


def play_two_humans(output: dict[str, Any]) -> bool:
    """
    Callback function for the "play_two_humans" button.
    Sets the output dictionary with the selected options for play_two_humans.

    Args:
        output: The output dictionary to store the selected options.

    Returns:
        True.
    """
    output["type"] = "play_two_humans"
    return True


def visualize_a_tree(output: dict[str, Any]) -> bool:
    """
    Callback function for the "Visualize a tree" button.
    Sets the output dictionary with the selected options for tree visualization.

    Args:
        output: The output dictionary to store the selected options.

    Returns:
        True.
    """
    output["type"] = "tree_visualization"
    return True



"""
This module contains the implementation of the NodeEvaluator class, which is responsible for evaluating the value of
 nodes in a tree-based move selector.

The NodeEvaluator class wraps a board evaluator and a syzygy table to provide more complex evaluations of chess
 positions. It handles queries for evaluating nodes and manages obvious over events.

Classes:
- NodeEvaluator: Wrapping node evaluator with syzygy and obvious over event.

Enums:
- NodeEvaluatorTypes: Types of node evaluators.

Constants:
- DISCOUNT: Discount factor used in the evaluation.

Functions:
- None

"""

from enum import Enum
from typing import Any

import chess

import chipiron.environments.chess.board as boards
import chipiron.players.boardevaluators as board_evals
from chipiron.players.boardevaluators.over_event import HowOver, OverEvent, Winner
from chipiron.players.boardevaluators.table_base.syzygy_table import SyzygyTable
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.algorithm_node import (
    AlgorithmNode,
)
from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode

DISCOUNT = 0.999999999999  # todo play with this


class NodeEvaluatorTypes(str, Enum):
    """
    Enum class representing different types of node evaluators.
    """

    NeuralNetwork = "neural_network"


class EvaluationQueries:
    """
    A class that represents evaluation queries for algorithm nodes.

    Attributes:
        over_nodes (list[AlgorithmNode]): A list of algorithm nodes that are considered "over".
        not_over_nodes (list[AlgorithmNode]): A list of algorithm nodes that are not considered "over".
    """

    over_nodes: list[AlgorithmNode]
    not_over_nodes: list[AlgorithmNode]

    def __init__(self) -> None:
        """
        Initializes a new instance of the NodeEvaluator class.
        """
        self.over_nodes = []
        self.not_over_nodes = []

    def clear_queries(self) -> None:
        """
        Clears the evaluation queries by resetting the over_nodes and not_over_nodes lists.
        """
        self.over_nodes = []
        self.not_over_nodes = []


class NodeEvaluator:
    """
    The NodeEvaluator class is responsible for evaluating the value of nodes in a tree structure.
    It uses a board evaluator and a syzygy evaluator to calculate the value of the nodes.
    """

    board_evaluator: board_evals.BoardEvaluator
    syzygy_evaluator: SyzygyTable[Any] | None

    def __init__(
        self,
        board_evaluator: board_evals.BoardEvaluator,
        syzygy: SyzygyTable[Any] | None,
    ) -> None:
        """
        Initializes a NodeEvaluator object.

        Args:
            board_evaluator (board_evals.BoardEvaluator): The board evaluator used to evaluate the chess board.
            syzygy (SyzygyTable | None): The Syzygy table used for endgame tablebase evaluations, or None if not available.
        """
        self.board_evaluator = board_evaluator
        self.syzygy_evaluator = syzygy

    def value_white(self, node: ITreeNode[Any]) -> float:
        """
        Calculates the value for the white player of a given node.
        If the value can be obtained from the syzygy evaluator, it is used.
        Otherwise, the board evaluator is used.
        """
        value_white: float | None = self.syzygy_value_white(node.board)
        value_white_float: float
        if value_white is None:
            value_white_float = self.board_evaluator.value_white(node.board)
        else:
            value_white_float = value_white
        return value_white_float

    def syzygy_value_white(self, board: boards.IBoard) -> float | None:
        """
        Calculates the value for the white player of a given board using the syzygy evaluator.
        If the syzygy evaluator is not available or the board is not in the syzygy table, None is returned.
        """
        if self.syzygy_evaluator is None or not self.syzygy_evaluator.fast_in_table(
            board
        ):
            return None
        else:
            val: int = self.syzygy_evaluator.val(board)
            return val

    def check_obvious_over_events(self, node: AlgorithmNode) -> None:
        """
        Updates the node.over object if the game is obviously over.
        """
        game_over: bool = node.tree_node.board.is_game_over()
        if game_over:
            value_as_string: str = node.board.result(claim_draw=True)
            how_over_: HowOver
            who_is_winner_: Winner
            match value_as_string:
                case "0-1":
                    how_over_ = HowOver.WIN
                    who_is_winner_ = Winner.BLACK
                case "1-0":
                    how_over_ = HowOver.WIN
                    who_is_winner_ = Winner.WHITE
                case "1/2-1/2":
                    how_over_ = HowOver.DRAW
                    who_is_winner_ = Winner.NO_KNOWN_WINNER
                case other:
                    raise ValueError(f"value {other} not expected in {__name__}")

            node.minmax_evaluation.over_event.becomes_over(
                how_over=how_over_,
                who_is_winner=who_is_winner_,
                termination=node.board.termination(),
            )

        elif self.syzygy_evaluator and self.syzygy_evaluator.fast_in_table(
            node.tree_node.board
        ):
            who_is_winner_, how_over_ = self.syzygy_evaluator.get_over_event(
                board=node.board
            )
            node.minmax_evaluation.over_event.becomes_over(
                how_over=how_over_,
                who_is_winner=who_is_winner_,
                termination=None,  # not sure how to retrieve this info more precisely atm
            )

    def value_white_from_over_event(
        self, over_event: OverEvent
    ) -> board_evals.ValueWhiteWhenOver:
        """
        Returns the value white given an over event.
        """
        assert over_event.is_over()
        if over_event.is_win():
            assert not over_event.is_draw()
            if over_event.is_winner(chess.WHITE):
                return board_evals.ValueWhiteWhenOver.VALUE_WHITE_WHEN_OVER_WHITE_WINS
            else:
                return board_evals.ValueWhiteWhenOver.VALUE_WHITE_WHEN_OVER_BLACK_WINS
        else:  # draw
            assert over_event.is_draw()
            return board_evals.ValueWhiteWhenOver.VALUE_WHITE_WHEN_OVER_DRAW

    def evaluate_over(self, node: AlgorithmNode) -> None:
        """
        Evaluates the node when the game is over.
        """
        evaluation = DISCOUNT**node.half_move * self.value_white_from_over_event(
            node.minmax_evaluation.over_event
        )
        node.minmax_evaluation.set_evaluation(evaluation)

    def evaluate_all_queried_nodes(self, evaluation_queries: EvaluationQueries) -> None:
        """
        Evaluates all the queried nodes.
        """
        node_over: AlgorithmNode
        for node_over in evaluation_queries.over_nodes:
            # assert isinstance(node_over, AlgorithmNode)
            self.evaluate_over(node_over)

        if evaluation_queries.not_over_nodes:
            self.evaluate_all_not_over(evaluation_queries.not_over_nodes)

        evaluation_queries.clear_queries()

    def add_evaluation_query(
        self, node: AlgorithmNode, evaluation_queries: EvaluationQueries
    ) -> None:
        """
        Adds an evaluation query for a node.
        """
        assert node.minmax_evaluation.value_white_evaluator is None
        self.check_obvious_over_events(node)
        if node.is_over():
            evaluation_queries.over_nodes.append(node)
        else:
            evaluation_queries.not_over_nodes.append(node)

    def evaluate_all_not_over(self, not_over_nodes: list[AlgorithmNode]) -> None:
        """
        Evaluates all the nodes that are not over.
        """

        node_not_over: AlgorithmNode
        for node_not_over in not_over_nodes:
            evaluation = self.value_white(node_not_over)
            processed_evaluation = self.process_evalution_not_over(
                evaluation=evaluation, node=node_not_over
            )
            # assert isinstance(node_not_over, AlgorithmNode)
            node_not_over.minmax_evaluation.set_evaluation(processed_evaluation)

    def process_evalution_not_over(
        self, evaluation: float, node: AlgorithmNode
    ) -> float:
        """
        Processes the evaluation for a node that is not over.
        """
        processed_evaluation = (1 / DISCOUNT) ** node.half_move * evaluation
        return processed_evaluation



"""
Module where we define the Stockfish Board Evaluator
"""

from dataclasses import dataclass

import chess.engine

import chipiron.environments.chess.board as boards


# TODO are we calling this?
@dataclass
class StockfishBoardEvalArgs:
    """Represents the arguments for the Stockfish board evaluator.

    Attributes:
        depth (int): The depth of the search algorithm.
        time_limit (float): The time limit for the search algorithm.
    """

    depth: int = 20
    time_limit: float = 0.1


class StockfishBoardEvaluator:
    """
    A board evaluator powered by stockfish
    """

    engine: chess.engine.SimpleEngine | None

    def __init__(self, args: StockfishBoardEvalArgs) -> None:
        """
        Initializes a StockfishBoardEvaluator object.

        Args:
            args (StockfishBoardEvalArgs): The arguments for the StockfishBoardEvaluator.

        Returns:
            None
        """
        self.engine = None

    def value_white(self, board: boards.IBoard) -> float:
        """
        Computes the value of the board for the white player.

        Args:
            board (boards.BoardChi): The board object representing the current state of the game.

        Returns:
            float: The value of the board for the white player.
        """
        # todo make a large reformat so that the players are created after the launch of process
        return 0.0  # to win time atm, please make it work fast when in testing mode and 0.1s in normal mode

        # if self.engine is None:
        #     # if this object is created in the init then seending the object
        #     self.engine = chess.engine.SimpleEngine.popen_uci(
        #         # TODO: should we remove the hardcoding
        #         r"stockfish/stockfish/stockfish-ubuntu-x86-64-avx2")
        # info = self.engine.analyse(board.board, chess.engine.Limit(time=0.01))
        # self.engine.quit()
        # self.engine = None
        # return info["score"]



"""
This module contains the implementation of the CommandLineHumanMoveSelector class, which allows a human player
 to select moves through the command line interface.
"""

from dataclasses import dataclass
from typing import Literal

import chipiron.environments.chess.board as boards
from chipiron.environments.chess.move import moveUci
from chipiron.environments.chess.move.imove import moveKey
from chipiron.utils import seed

from .move_selector import MoveRecommendation
from .move_selector_types import MoveSelectorTypes


@dataclass
class CommandLineHumanPlayerArgs:
    """
    Represents the arguments for a human player that selects moves through the command line interface.
    """

    type: Literal[MoveSelectorTypes.CommandLineHuman]  # for serialization


@dataclass
class GuiHumanPlayerArgs:
    """
    Represents the arguments for a human player that selects moves through the GUI.
    """

    type: Literal[MoveSelectorTypes.GuiHuman]  # for serialization


class CommandLineHumanMoveSelector:
    """
    A move selector that allows a human player to select moves through the command line interface.
    """

    def select_move(self, board: boards.IBoard, move_seed: seed) -> MoveRecommendation:
        """
        Selects a move based on user input through the command line interface.

        Args:
            board (boards.BoardChi): The current state of the chess board.
            move_seed (seed): The seed used for move selection.

        Returns:
            MoveRecommendation: The selected move recommendation.

        Raises:
            AssertionError: If the selected move is not a legal move.
        """
        print(f"Legal Moves {board.legal_moves}")
        legal_moves_uci: list[moveUci] = []
        move: moveKey
        for move in board.legal_moves:
            legal_moves_uci.append(board.get_uci_from_move_key(move))
        print(f"Legal Moves {legal_moves_uci}")

        good_move: bool = False
        while not good_move:
            move_uci: str = input("Input your move")
            print(f"choice of move {move_uci}")
            if move_uci in legal_moves_uci:
                good_move = True
            else:
                print("Bad move, Not legal")

        assert move_uci in legal_moves_uci

        return MoveRecommendation(move=move_uci, evaluation=None)



"""
This module defines the AlgorithmNode class, which is a generic node used by the tree and value algorithm.
It wraps tree nodes with values, minimax computation, and exploration tools.
"""

from typing import Any

import chess

import chipiron.environments.chess.board as boards
from chipiron.environments.chess.board.iboard import LegalMoveKeyGeneratorP
from chipiron.environments.chess.move.imove import moveKey
from chipiron.players.boardevaluators.neural_networks.input_converters.board_representation import (
    BoardRepresentation,
)
from chipiron.players.move_selector.treevalue.indices.node_indices import (
    NodeExplorationData,
)
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.node_minmax_evaluation import (
    NodeMinmaxEvaluation,
)
from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode
from chipiron.players.move_selector.treevalue.nodes.tree_node import TreeNode


class AlgorithmNode:
    """
    The generic Node used by the tree and value algorithm.
    It wraps tree nodes with values, minimax computation and exploration tools
    """

    tree_node: TreeNode[Any]  # the reference to the tree node that is wrapped
    minmax_evaluation: NodeMinmaxEvaluation  # the object computing the value
    exploration_index_data: (
        NodeExplorationData | None
    )  # the object storing the information to help the algorithm decide the next nodes to explore
    board_representation: BoardRepresentation | None  # the board representation

    def __init__(
        self,
        tree_node: TreeNode[Any],
        minmax_evaluation: NodeMinmaxEvaluation,
        exploration_index_data: NodeExplorationData | None,
        board_representation: BoardRepresentation | None,
    ) -> None:
        """
        Initializes an AlgorithmNode object.

        Args:
            tree_node (TreeNode): The tree node that is wrapped.
            minmax_evaluation (NodeMinmaxEvaluation): The object computing the value.
            exploration_index_data (NodeExplorationData | None): The object storing the information to help the algorithm decide the next nodes to explore.
            board_representation (BoardRepresentation | None): The board representation.
        """
        self.tree_node = tree_node
        self.minmax_evaluation = minmax_evaluation
        self.exploration_index_data = exploration_index_data
        self.board_representation = board_representation

    @property
    def player_to_move(self) -> chess.Color:
        """
        Returns the color of the player to move.

        Returns:
            chess.Color: The color of the player to move.
        """
        return self.tree_node.player_to_move

    @property
    def id(self) -> int:
        """
        Returns the ID of the node.

        Returns:
            int: The ID of the node.
        """
        return self.tree_node.id

    @property
    def half_move(self) -> int:
        """
        Returns the half move count.

        Returns:
            int: The half move count.
        """
        return self.tree_node.half_move

    @property
    def fast_rep(self) -> boards.boardKey:
        """
        Returns the fast representation of the node.

        Returns:
            str: The fast representation of the node.
        """
        return self.tree_node.fast_rep

    @property
    def moves_children(self) -> dict[moveKey, ITreeNode[Any] | None]:
        """
        Returns the bidirectional dictionary of moves and their corresponding child nodes.

        Returns:
            dict[IMove, ITreeNode[Any] | None]: The bidirectional dictionary of moves and their corresponding child nodes.
        """
        return self.tree_node.moves_children

    @property
    def parent_nodes(self) -> dict[ITreeNode[Any], moveKey]:
        """
        Returns the dictionary of parent nodes of the current tree node with associated move.

        :return: A dictionary of parent nodes of the current tree node with associated move.
        """
        return self.tree_node.parent_nodes

    @property
    def board(self) -> boards.IBoard:
        """
        Returns the board.

        Returns:
            BoardChi: The board.
        """
        return self.tree_node.board

    def is_over(self) -> bool:
        """
        Checks if the game is over.

        Returns:
            bool: True if the game is over, False otherwise.
        """
        return self.minmax_evaluation.is_over()

    def add_parent(self, move: moveKey, new_parent_node: ITreeNode[Any]) -> None:
        """
        Adds a parent node.

        Args:
            move (IMove): the move that led to the node from the new_parent_node
            new_parent_node (ITreeNode): The new parent node to add.
        """
        self.tree_node.add_parent(move=move, new_parent_node=new_parent_node)

    @property
    def legal_moves(self) -> LegalMoveKeyGeneratorP:
        """
        Returns the legal move generator.

        Returns:
            chess.LegalMoveGenerator: The legal move generator.
        """
        return self.tree_node.board_.legal_moves

    @property
    def all_legal_moves_generated(self) -> bool:
        """
        Returns True if all legal moves have been generated, False otherwise.

        Returns:
            bool: True if all legal moves have been generated, False otherwise.
        """
        return self.tree_node.all_legal_moves_generated

    @all_legal_moves_generated.setter
    def all_legal_moves_generated(self, value: bool) -> None:
        """
        Sets the flag indicating if all legal moves have been generated.

        Args:
            value (bool): The value to set.
        """
        self.tree_node.all_legal_moves_generated = value

    @property
    def non_opened_legal_moves(self) -> set[moveKey]:
        """
        Returns the set of non-opened legal moves.

        Returns:
            set[IMove]: The set of non-opened legal moves.
        """
        return self.tree_node.non_opened_legal_moves

    def dot_description(self) -> str:
        """
        Returns the dot description of the node.

        Returns:
            str: The dot description of the node.
        """
        exploration_description: str = (
            self.exploration_index_data.dot_description()
            if self.exploration_index_data is not None
            else ""
        )

        return f"{self.tree_node.dot_description()}\n{self.minmax_evaluation.dot_description()}\n{exploration_description}"

    def __str__(self) -> str:
        return f"{self.__class__} id :{self.tree_node.id}"



ML_FLOW_URI_PATH = (
    "sqlite:///chipiron/scripts/default_output_folder/mlflow_data/mlruns.db"
)

ML_FLOW_URI_PATH_TEST = (
    "sqlite:///chipiron/scripts/default_output_folder/mlflow_data/mlruns_test.db"
)



from abc import abstractmethod
from typing import Protocol

from .utils import moveUci

# numbering scheme for actions in the node of the trees
moveKey = int


class IMove(Protocol):

    @abstractmethod
    def uci(self) -> moveUci: ...



"""
Module for the ChiNN class
"""

import sys

import torch
import torch.nn as nn

from chipiron.utils import path
from chipiron.utils.logger import chipiron_logger


class ChiNN(nn.Module):
    """
    The Generic Neural network class of chipiron
    """

    def __init__(self) -> None:
        """
        Initializes an instance of the ChiNN class.
        """
        super(ChiNN, self).__init__()

    def __getstate__(self) -> None:
        """
        Get the state of the object for pickling.

        Returns:
            None
        """
        return None

    def init_weights(self) -> None:
        """
        Initialize the weights of the model.
        """
        pass

    def load_weights_from_file(self, path_to_param_file: path) -> None:
        """
        Loads the neural network weights from a file or initializes them if the file doesn't exist.

        Args:
            path_to_param_file (str): The path to the parameter file.
            authorisation_to_create_file (bool): Flag indicating whether the program has authorization to create a new file.

        Returns:
            None
        """
        chipiron_logger.info(f"load_or_init_weights from {path_to_param_file}")
        try:  # load
            with open(path_to_param_file, "rb") as fileNNR:
                chipiron_logger.info(
                    f"loading the existing param file {path_to_param_file}"
                )
                if torch.cuda.is_available():
                    self.load_state_dict(torch.load(fileNNR))
                else:
                    self.load_state_dict(
                        torch.load(fileNNR, map_location=torch.device("cpu"))
                    )

        except EnvironmentError:  # init
            chipiron_logger.error(f"no file {path_to_param_file}")
            sys.exit(
                "Error: no NN weights file and no rights to create it for file {}".format(
                    path_to_param_file,
                )
            )

    def log_readable_model_weights_to_file(self, file_path: str) -> None:
        raise Exception("not implemented in base class")



import os
from enum import Enum

import parsley_coco

from chipiron.players import PlayerArgs
from chipiron.utils import path


class PlayerConfigTag(str, Enum):
    """
    This class is used to identify the player configuration files.
    Each player configuration file should be listed here as a class attribute.
    The class also provides methods to check if a player is human and to get the YAML file path for the player configuration.
    """

    CHIPIRON = "Chipiron"

    GUI_HUMAN = "Gui_Human"
    COMMAND_LINE_HUMAN = "Command_Line_Human"
    SEQUOOL = "Sequool"
    RECUR_ZIPF_BASE_3 = "RecurZipfBase3"
    RECUR_ZIPF_BASE_4 = "RecurZipfBase4"
    UNIFORM_DEPTH = "UniformDepth"
    UNIFORM_DEPTH_2 = "UniformDepth2"
    UNIFORM_DEPTH_3 = "UniformDepth3"
    UNIFORM = "Uniform"
    RANDOM = "Random"
    Stockfish = "Stockfish"

    Test_Sequool = "players_for_test_purposes/Sequool"
    Test_RecurZipfSequool = "players_for_test_purposes/RecurZipfSequool"
    Test_RecurZipfBase3 = "players_for_test_purposes/RecurZipfBase3"
    Test_RecurZipfBase4 = "players_for_test_purposes/RecurZipfBase4"
    Test_Uniform = "players_for_test_purposes/Uniform"

    def is_human(self) -> bool:
        """Check if the player is human.
        This method checks if the player is a human player based on the player's configuration tag.
        It returns True if the player is a human player, and False otherwise.

        Returns:
            bool: True if the player is a human player, False otherwise.
        """
        return (
            self is PlayerConfigTag.GUI_HUMAN
            or self is PlayerConfigTag.COMMAND_LINE_HUMAN
        )

    def get_yaml_file_path(self) -> path:
        """Get the YAML file path for the player configuration.
        This method constructs the file path for the player configuration YAML file
        based on the player's configuration tag.
        It returns the file path as a string.

        Returns:
            path: The file path for the player configuration YAML file.
        Raises:
            ValueError: If the player configuration tag is not recognized.
        """
        path_player: path
        if self is PlayerConfigTag.CHIPIRON:
            path_player = os.path.join(
                "data/players/player_config/chipiron/chipiron.yaml"
            )
        else:
            path_player = os.path.join(
                "data/players/player_config", self.value + ".yaml"
            )
        return path_player

    def get_players_args(self) -> PlayerArgs:
        """Get the player arguments from the YAML file.
        This method fetches the player arguments from the YAML file
        corresponding to the player's configuration tag.

        Returns:
            PlayerArgs: The player arguments as a dataclass.
        """
        player_args: PlayerArgs = parsley_coco.resolve_yaml_file_to_base_dataclass(
            yaml_path=self.get_yaml_file_path(),
            base_cls=PlayerArgs,
            raise_error_with_nones=False,
        )
        return player_args


if __name__ == "__main__":
    a = PlayerConfigTag.COMMAND_LINE_HUMAN



"""
Module for creating neural networks and neural network board evaluators.
"""

import os.path
import sys
from dataclasses import dataclass
from enum import Enum
from typing import Any

import dacite
from sympy import N

from chipiron.players.boardevaluators.neural_networks.input_converters.board_to_input import (
    BoardToInputFunction,
    create_board_to_input,
)
from chipiron.players.boardevaluators.neural_networks.models.multi_layer_perceptron import (
    MultiLayerPerceptron,
    MultiLayerPerceptronArgs,
)
from chipiron.players.boardevaluators.neural_networks.models.transformer_one import (
    TransformerArgs,
    TransformerOne,
)
from chipiron.players.boardevaluators.neural_networks.neural_net_board_eval_args import (
    NeuralNetArchitectureArgs,
)
from chipiron.players.boardevaluators.neural_networks.nn_board_evaluator import (
    NNBoardEvaluator,
)
from chipiron.players.boardevaluators.neural_networks.NNModelType import NNModelType
from chipiron.players.boardevaluators.neural_networks.NNModelTypeArgs import (
    NNModelTypeArgs,
)
from chipiron.players.boardevaluators.neural_networks.output_converters.factory import (
    create_output_converter,
)
from chipiron.players.boardevaluators.neural_networks.output_converters.output_value_converter import (
    OutputValueConverter,
)
from chipiron.utils import path, yaml_fetch_args_in_file
from chipiron.utils.chi_nn import ChiNN


@dataclass
class NeuralNetModelsAndArchitecture:
    """
    Class to hold the neural network models and architecture.
    Attributes:
        model_weights_file_name (path): The file name of the model weights.
        nn_architecture_args (NeuralNetArchitectureArgs): The neural network architecture arguments.
    """

    model_weights_file_name: path
    nn_architecture_args: NeuralNetArchitectureArgs

    @classmethod
    def build_from_folder_path(
        cls, folder_path: path
    ) -> "NeuralNetModelsAndArchitecture":
        """
        Build an instance of NeuralNetModelsAndArchitecture from the given folder path.

        Args:
            folder_path (Path): Path to the folder containing 'architecture.yaml' and model weights.

        Returns:
            NeuralNetModelsAndArchitecture: An initialized instance.
        """
        nn_args = get_architecture_args_from_folder(folder_path=folder_path)
        model_file = os.path.join(folder_path, nn_args.filename() + ".pt")

        return cls(model_weights_file_name=model_file, nn_architecture_args=nn_args)


def get_nn_param_file_path_from(
    folder_path: path, file_name: str | None = None
) -> tuple[str, str]:
    """
    Get the file path for the neural network parameters.

    Args:
        folder_path (str): The folder path for the neural network parameters.

    Returns:
        str: The file path for the neural network parameters.
    """
    nn_param_file_path: str
    if file_name is None:
        nn_param_file_path = os.path.join(folder_path, "param")
    else:
        nn_param_file_path = os.path.join(folder_path, file_name)
    return nn_param_file_path + ".pt", nn_param_file_path + ".yaml"


def get_nn_architecture_file_path_from(folder_path: path) -> str:
    """
    Get the file path for the architecture parameters.

    Args:
        folder_path (str): The folder path for the architecture parameters.

    Returns:
        str: The file path for the architecture parameters.
    """
    nn_param_file_path: str = os.path.join(folder_path, "architecture.yaml")
    return nn_param_file_path


def create_nn(nn_type_args: NNModelTypeArgs) -> ChiNN:
    """
    Create a neural network.
    """

    net: ChiNN
    match nn_type_args:
        case MultiLayerPerceptronArgs():
            net = MultiLayerPerceptron(args=nn_type_args)
        case TransformerArgs():
            net = TransformerOne(args=nn_type_args)
        case other:
            sys.exit(f"Create NN: can not find {other} in file {__name__}")
    return net


# todo: probably dead code, check!
def get_architecture_args_from_file(
    architecture_file_name: path,
) -> NeuralNetArchitectureArgs:
    args_dict: dict[Any, Any] = yaml_fetch_args_in_file(
        path_file=architecture_file_name
    )
    nn_architecture_args: NeuralNetArchitectureArgs = dacite.from_dict(
        data_class=NeuralNetArchitectureArgs,
        data=args_dict,
        config=dacite.Config(cast=[Enum]),
    )
    return nn_architecture_args


def get_architecture_args_from_folder(folder_path: path) -> NeuralNetArchitectureArgs:
    architecture_file_name: path = get_nn_architecture_file_path_from(
        folder_path=folder_path
    )
    if not os.path.isfile(architecture_file_name):
        raise Exception(f"this is not a file {architecture_file_name}")

    nn_architecture_args: NeuralNetArchitectureArgs = get_architecture_args_from_file(
        architecture_file_name=architecture_file_name
    )

    return nn_architecture_args


def create_nn_from_param_path_and_architecture_args(
    model_weights_file_name: path, nn_architecture_args: NeuralNetArchitectureArgs
) -> tuple[ChiNN, NeuralNetArchitectureArgs]:
    net: ChiNN = create_nn(nn_type_args=nn_architecture_args.model_type_args)
    net.load_weights_from_file(path_to_param_file=model_weights_file_name)
    return net, nn_architecture_args


def create_nn_from_folder_path_and_existing_model(
    folder_path: path,
) -> tuple[ChiNN, NeuralNetArchitectureArgs]:
    nn_architecture_args: NeuralNetArchitectureArgs = get_architecture_args_from_folder(
        folder_path=folder_path
    )
    model_weights_file_name: path = os.path.join(folder_path, "param.pt")

    return create_nn_from_param_path_and_architecture_args(
        model_weights_file_name=model_weights_file_name,
        nn_architecture_args=nn_architecture_args,
    )


def create_nn_board_eval_from_folder_path_and_existing_model(
    path_to_nn_folder: path,
) -> tuple[NNBoardEvaluator, NeuralNetArchitectureArgs]:
    """
    Create a neural network board evaluator.

    Args:
        path_to_nn_folder (path): the path to the folder where the model is defined.

    Returns:
        NNBoardEvaluator: The created neural network board evaluator.
    """
    net: ChiNN
    nn_architecture_args: NeuralNetArchitectureArgs
    net, nn_architecture_args = create_nn_from_folder_path_and_existing_model(
        folder_path=path_to_nn_folder
    )

    nn_board_evaluator = create_nn_board_eval_from_nn_and_architecture_args(
        nn=net, nn_architecture_args=nn_architecture_args
    )
    return nn_board_evaluator, nn_architecture_args


def create_nn_board_eval_from_nn_and_architecture_args(
    nn_architecture_args: NeuralNetArchitectureArgs,
    nn: ChiNN,
) -> NNBoardEvaluator:
    board_to_input_convert: BoardToInputFunction = create_board_to_input(
        model_input_representation_type=nn_architecture_args.model_input_representation_type
    )

    output_and_value_converter: OutputValueConverter = create_output_converter(
        model_output_type=nn_architecture_args.model_output_type
    )

    return NNBoardEvaluator(
        net=nn,
        output_and_value_converter=output_and_value_converter,
        board_to_input_convert=board_to_input_convert,
    )


def create_nn_board_eval_from_architecture_args(
    nn_architecture_args: NeuralNetArchitectureArgs,
) -> NNBoardEvaluator:
    nn = create_nn(nn_type_args=nn_architecture_args.model_type_args)
    nn.init_weights()

    return create_nn_board_eval_from_nn_and_architecture_args(
        nn_architecture_args=nn_architecture_args, nn=nn
    )


def create_nn_board_eval_from_nn_parameters_file_and_existing_model(
    model_weights_file_name: path, nn_architecture_args: NeuralNetArchitectureArgs
) -> NNBoardEvaluator:
    net: ChiNN
    net, nn_architecture_args = create_nn_from_param_path_and_architecture_args(
        model_weights_file_name=model_weights_file_name,
        nn_architecture_args=nn_architecture_args,
    )

    return create_nn_board_eval_from_nn_and_architecture_args(
        nn_architecture_args=nn_architecture_args, nn=net
    )



"""
This module defines the ScriptType enum and provides a mapping between script types and their corresponding script classes.
"""

from enum import Enum


class ScriptType(str, Enum):
    """
    Enum class representing different types of scripts.
    """

    OneMatch = "one_match"
    League = "league"
    LearnNN = "learn_nn"
    LearnNNFromScratch = "learn_nn_from_scratch"
    BaseTreeExploration = "base_tree_exploration"
    TreeVisualization = "tree_visualization"
    ReplayMatch = "replay_match"



"""
This module contains functions for creating match managers in the Chipiron game framework.
"""

import multiprocessing
import queue
from typing import Any

import chipiron as ch
import chipiron.games.game as game
import chipiron.players as players
from chipiron.environments.chess.board.factory import BoardFactory, create_board_factory
from chipiron.games.game.game_manager_factory import GameManagerFactory
from chipiron.games.match.match_args import MatchArgs
from chipiron.games.match.match_manager import MatchManager
from chipiron.games.match.match_results_factory import MatchResultsFactory
from chipiron.games.match.utils import fetch_match_games_args_convert_and_save
from chipiron.players.boardevaluators.board_evaluator import IGameBoardEvaluator
from chipiron.players.boardevaluators.factory import create_game_board_evaluator
from chipiron.players.boardevaluators.table_base.factory import create_syzygy
from chipiron.scripts.chipiron_args import ImplementationArgs
from chipiron.scripts.script_args import BaseScriptArgs
from chipiron.utils import path
from chipiron.utils.dataclass import IsDataclass

from ...environments.chess.move_factory import MoveFactory, create_move_factory
from ...players.boardevaluators.table_base import SyzygyTable
from .match_settings_args import MatchSettingsArgs


def create_match_manager(
    args_match: MatchSettingsArgs,
    args_player_one: players.PlayerArgs,
    args_player_two: players.PlayerArgs,
    args_game: game.GameArgs,
    implementation_args: ImplementationArgs,
    universal_behavior: bool = False,
    seed: int | None = None,
    output_folder_path: path | None = None,
    gui: bool = False,
) -> MatchManager:
    """
    Create a match manager for running matches between two players.

    Args:
        implementation_args: (ImplementationArgs) the implementation args
        args_match (MatchSettingsArgs): The match settings arguments.
        args_player_one (players.PlayerArgs): The arguments for player one.
        args_player_two (players.PlayerArgs): The arguments for player two.
        args_game (game.GameArgs): The game arguments.
        seed (int | None, optional): The seed for random number generation. Defaults to None.
        output_folder_path (path | None, optional): The output folder path. Defaults to None.
        gui (bool, optional): Flag indicating whether to enable GUI. Defaults to False.

    Returns:
        MatchManager: The created match manager.
    """
    main_thread_mailbox: queue.Queue[IsDataclass] = multiprocessing.Manager().Queue()

    # Creation of the Syzygy table for perfect play in low pieces cases, needed by the GameManager
    # and can also be used by the players
    syzygy_table: SyzygyTable[Any] | None = create_syzygy(
        use_rust=implementation_args.use_rust_boards,
    )

    player_one_name: str = args_player_one.name
    player_two_name: str = args_player_two.name

    can_stockfish: bool = args_player_one.name not in [
        "Stockfishu"
    ] and args_player_two.name not in ["Stockfishu"]
    game_board_evaluator: IGameBoardEvaluator = create_game_board_evaluator(
        gui=gui, can_stockfish=can_stockfish
    )

    board_factory: BoardFactory = create_board_factory(
        use_rust_boards=implementation_args.use_rust_boards,
        use_board_modification=implementation_args.use_board_modification,
        sort_legal_moves=universal_behavior,
    )

    move_factory: MoveFactory = create_move_factory(
        use_rust_boards=implementation_args.use_rust_boards
    )

    game_manager_factory: GameManagerFactory = GameManagerFactory(
        syzygy_table=syzygy_table,
        game_manager_board_evaluator=game_board_evaluator,
        output_folder_path=output_folder_path,
        main_thread_mailbox=main_thread_mailbox,
        board_factory=board_factory,
        implementation_args=implementation_args,
        move_factory=move_factory,
        universal_behavior=universal_behavior,
    )

    match_results_factory: MatchResultsFactory = MatchResultsFactory(
        player_one_name=player_one_name, player_two_name=player_two_name
    )

    game_args_factory: game.GameArgsFactory = game.GameArgsFactory(
        args_match=args_match,
        args_player_one=args_player_one,
        args_player_two=args_player_two,
        seed_=seed,
        args_game=args_game,
    )

    match_manager: MatchManager = MatchManager(
        player_one_id=player_one_name,
        player_two_id=player_two_name,
        game_manager_factory=game_manager_factory,
        game_args_factory=game_args_factory,
        match_results_factory=match_results_factory,
        output_folder_path=output_folder_path,
    )
    return match_manager


def create_match_manager_from_args(
    match_args: MatchArgs,
    base_script_args: BaseScriptArgs,
    implementation_args: ImplementationArgs,
    gui: bool = False,
) -> MatchManager:
    """
    Create a match manager from the given arguments.

    Args:
        implementation_args(ImplementationArgs): The implementation args
        match_args (MatchArgs): The match arguments.
        base_script_args (ScriptArgs) The script arguments.
        gui (bool, optional): Flag indicating whether to enable GUI. Defaults to False.

    Returns:
        MatchManager: The created match manager.
    """

    assert isinstance(match_args.player_one, players.PlayerArgs)
    assert isinstance(match_args.player_two, players.PlayerArgs)

    player_one_args: players.PlayerArgs = match_args.player_one
    player_two_args: players.PlayerArgs = match_args.player_two

    # Recovering args from yaml file for match and game and merging with extra args and converting
    # to standardized dataclass
    match_setting_args: MatchSettingsArgs
    game_args: game.GameArgs
    match_setting_args, game_args = fetch_match_games_args_convert_and_save(
        profiling=base_script_args.profiling,
        testing=base_script_args.testing,
        match_args=match_args,
        experiment_output_folder=base_script_args.experiment_output_folder,
    )

    assert (
        player_one_args.name != "Command_Line_Human.yaml"
        or not game_args.each_player_has_its_own_thread
    )
    assert (
        player_two_args.name != "Command_Line_Human.yaml"
        or not game_args.each_player_has_its_own_thread
    )

    # taking care of random
    ch.set_seeds(seed=base_script_args.seed)

    match_manager: MatchManager = create_match_manager(
        args_match=match_setting_args,
        args_player_one=player_one_args,
        args_player_two=player_two_args,
        output_folder_path=base_script_args.experiment_output_folder,
        universal_behavior=base_script_args.universal_behavior,
        seed=base_script_args.seed,
        args_game=game_args,
        gui=gui,
        implementation_args=implementation_args,
    )

    return match_manager



from typing import Protocol

import chess
import shakmaty_python_binding

from chipiron.environments.chess.board.iboard import IBoard
from chipiron.environments.chess.move.imove import IMove

from .board import RustyBoardChi
from .move import moveUci


class MoveFactory(Protocol):
    def __call__(self, move_uci: moveUci, board: IBoard | None = None) -> IMove: ...


def create_move_factory(
    use_rust_boards: bool,
) -> MoveFactory:
    move_factory: MoveFactory

    if use_rust_boards:
        # todo can we go back to the not test version without the assert? generics or just typos?
        move_factory = create_rust_move_test_2
    else:
        move_factory = create_move
    return move_factory


def create_rust_move(
    move_uci: moveUci, board: RustyBoardChi | None = None
) -> shakmaty_python_binding.MyMove:
    assert board is not None
    return shakmaty_python_binding.MyMove(move_uci, board.chess_)


def create_rust_move_test_2(
    move_uci: moveUci, board: IBoard | None = None
) -> shakmaty_python_binding.MyMove:
    assert isinstance(board, RustyBoardChi)
    binding_move = shakmaty_python_binding.MyMove(move_uci, board.chess_)
    # rust_move = RustMove(move=binding_move, uci=move_uci)
    # return rust_move
    return binding_move


def create_rust_move_test(move_uci: moveUci, board: IBoard | None = None) -> chess.Move:
    assert board is not None
    return chess.Move.from_uci(move_uci)


def create_move(move_uci: moveUci, board: IBoard | None = None) -> chess.Move:
    return chess.Move.from_uci(move_uci)



"""
This module contains classes for handling datasets used in board evaluation tasks.

Classes:
- MyDataSet: A custom dataset class that loads and preprocesses data.
- FenAndValueDataSet: A subclass of MyDataSet that processes raw rows into input and target value tensors.

Functions:
- process_stockfish_value: A function that processes the stockfish value for a given board and row.
"""

import time
from dataclasses import dataclass
from typing import Any, Callable

import numpy as np
import pandas
import torch
from torch.utils.data import Dataset

from chipiron.environments.chess import BoardChi
from chipiron.environments.chess.board import IBoard
from chipiron.environments.chess.board.factory import create_board_chi
from chipiron.environments.chess.board.utils import FenPlusHistory, fen
from chipiron.players.boardevaluators.neural_networks.input_converters.board_to_input import (
    BoardToInputFunction,
)
from chipiron.utils import path
from chipiron.utils.logger import chipiron_logger


@dataclass
class DataSetArgs:
    train_file_name: path
    test_file_name: path | None = None
    preprocessing_data_set: bool = False


class MyDataSet(Dataset[Any]):
    """
    A custom dataset class that loads and preprocesses data.

    Attributes:
    - file_name (str): The file name of the dataset.
    - preprocessing (bool): Flag indicating whether to preprocess the dataset.
    - data (pandas.DataFrame | list[tuple[torch.Tensor, torch.Tensor]] | None): The loaded and processed data.
    - len (int | None): The length of the dataset.

    Methods:
    - load(): Loads the dataset from the file.
    - process_raw_row(row: pandas.Series) -> tuple[torch.Tensor, torch.Tensor]: Processes a raw row into input and target tensors.
    """

    data: pandas.DataFrame | list[tuple[torch.Tensor, torch.Tensor]] | None
    len: int | None

    def __init__(self, file_name: path, preprocessing: bool) -> None:
        """
        Initializes a new instance of the MyDataSet class.

        Args:
        - file_name (str): The file name of the dataset.
        - preprocessing (bool): Flag indicating whether to preprocess the dataset.
        """
        self.file_name = file_name
        self.preprocessing = preprocessing
        self.data = None
        self.len = None

    def load(self) -> None:
        """
        Loads the dataset from the file.
        """
        chipiron_logger.info("Loading the dataset...")
        start_time = time.time()
        raw_data: pandas.DataFrame = pandas.read_pickle(self.file_name)
        raw_data = (
            raw_data.copy()
        )  # gets rid of compatibility problems between various version of panda and pickle
        chipiron_logger.info(f"raw_data {type(raw_data)}")
        self.raw_data = raw_data

        chipiron_logger.info(
            "--- LOAD READ PICKLE %s seconds ---" % (time.time() - start_time)
        )
        chipiron_logger.info("Dataset  loaded with {} items".format(len(raw_data)))

        # preprocessing
        if self.preprocessing:
            chipiron_logger.info("preprocessing dataset...")
            processed_data: list[tuple[torch.Tensor, torch.Tensor]] = []

            for idx in range(len(raw_data)):
                # print(idx, type(idx),idx % 10 == 0)
                if idx % 10 == 0:
                    chipiron_logger.info(
                        f"\rloading the data {str(idx / len(raw_data) * 100)}%"
                    )
                row: pandas.Series = raw_data.iloc[idx % len(raw_data)]
                processed_data.append(self.process_raw_row(row))
            self.data = processed_data

            chipiron_logger.info("preprocessing dataset done")
        else:
            chipiron_logger.info("no preprocessing the dataset")
            self.data = raw_data

        self.data = (
            self.data.copy()
        )  # gets rid of compatibility problems between various version of panda and pickle

        self.len = len(self.data)

    def process_raw_row(self, row: pandas.Series) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Processes a raw row into input and target tensors.

        Args:
        - row (pandas.Series): The raw row from the dataset.

        Returns:
        - tuple[torch.Tensor, torch.Tensor]: The input and target tensors.
        """
        raise Exception("should not be called")

    def __len__(self) -> int:
        """
        Returns the length of the dataset.

        Returns:
        - int: The length of the dataset.
        """
        assert self.data is not None
        return len(self.data)

    def __getitem__(
        self, idx: int
    ) -> tuple[torch.Tensor, torch.Tensor] | pandas.Series:
        """
        Returns the item at the given index.

        Args:
        - idx (int): The index of the item.

        Returns:
        - tuple[torch.Tensor, torch.Tensor] | pandas.Series: The input and target tensors, or the raw row.
        """
        assert self.data is not None

        if self.preprocessing:
            assert self.len is not None
            return self.data[idx % self.len]
        else:
            assert isinstance(self.data, pandas.DataFrame)
            assert self.len is not None

            raw_row = self.data.iloc[idx % self.len]
            return self.process_raw_row(raw_row)  # to be coded!

    def get_unprocessed(self, idx: int) -> pandas.Series:
        """ """
        assert isinstance(self.data, pandas.DataFrame)
        assert self.len is not None

        raw_row = self.data.iloc[idx % self.len]
        return raw_row


def process_stockfish_value(row: pandas.Series) -> float:
    """
    Processes the stockfish value for a given board and row.

    Args:
    - board (BoardChi): The chess board.
    - row (pandas.Series): The row from the dataset.

    Returns:
    - torch.Tensor: The processed target value tensor.
    """
    # target values are value between -1 and 1 from the point of view of white. (+1 is white win and -1 is white loose)
    target_value: float = np.tanh(row["stockfish_value"] / 500.0)
    return target_value


class FenAndValueDataSet(MyDataSet):
    """
    A subclass of MyDataSet that processes raw rows into input and target value tensors.

    Attributes:
    - transform_board_function (BoardToInputFunction): The function to transform the board into input tensor.
    - transform_value_function (callable): The function to transform the value for a given board and row.

    Methods:
    - process_raw_row(row: pandas.Series) -> tuple[torch.Tensor, torch.Tensor]: Processes a raw row into input and target tensors.
    - process_raw_rows(dataframe: pandas.DataFrame) -> list[tuple[torch.Tensor, torch.Tensor]]: Processes raw rows into input and target tensors.
    """

    transform_board_function: BoardToInputFunction  # transform board to model input
    transform_dataset_value_to_white_value_function: Callable[
        [pandas.Series], float
    ]  # transform value in dataset to standardized value white float
    transform_white_value_to_model_output_function: Callable[
        [float, IBoard], torch.Tensor
    ]  # transform white value to model output

    def __init__(
        self,
        file_name: path,
        transform_white_value_to_model_output_function: Callable[
            [float, IBoard], torch.Tensor
        ],
        transform_dataset_value_to_white_value_function: Callable[
            [pandas.Series], float
        ],
        preprocessing: bool = False,
        transform_board_function: str | BoardToInputFunction = "identity",
    ) -> None:
        """
        Initializes a new instance of the FenAndValueDataSet class.

        Args:
        - file_name (str): The file name of the dataset.
        - preprocessing (bool): Flag indicating whether to preprocess the dataset.
        - transform_board_function (str | BoardToInputFunction): The function to transform the board into input tensor.
        - transform_value_function (str): The function to transform the value for a given board and row.
        """
        super().__init__(file_name, preprocessing)
        # transform function
        if transform_board_function == "identity":
            raise Exception(f"tobe coded in {__name__}")
        else:
            assert isinstance(transform_board_function, BoardToInputFunction)
            self.transform_board_function = transform_board_function

        # transform function
        self.transform_dataset_value_to_white_value_function = (
            transform_dataset_value_to_white_value_function
        )

        self.transform_white_value_to_model_output_function = (
            transform_white_value_to_model_output_function
        )

    def process_raw_row(self, row: pandas.Series) -> tuple[Any, torch.Tensor]:
        """
        Processes a raw row into input and target tensors.

        Args:
        - row (pandas.Series): The raw row from the dataset.

        Returns:
        - tuple[torch.Tensor, torch.Tensor]: The input and target tensors.
        """
        fen_: fen = row["fen"]

        # todo should we make it general and allow rust boards just for testing all comptabilities ?
        board: BoardChi = create_board_chi(
            fen_with_history=FenPlusHistory(current_fen=fen_),
            use_board_modification=True,
        )
        input_layer = self.transform_board_function(board)
        target_value_white: float = (
            self.transform_dataset_value_to_white_value_function(row)
        )
        target_value: torch.Tensor = (
            self.transform_white_value_to_model_output_function(
                target_value_white, board
            )
        )
        return input_layer, target_value

    def process_raw_rows(
        self, dataframe: pandas.DataFrame
    ) -> list[tuple[torch.Tensor, torch.Tensor]]:
        """
        Processes raw rows into input and target tensors.

        Args:
        - dataframe (pandas.DataFrame): The raw rows from the dataset.

        Returns:
        - list[tuple[torch.Tensor, torch.Tensor]]: The processed input and target tensors.
        """
        processed_data: list[tuple[torch.Tensor, torch.Tensor]] = []
        row: pandas.Series
        for row in dataframe.iloc:
            processed_data.append(self.process_raw_row(row))

        return processed_data



"""
the one match script
"""

import multiprocessing
import os
import queue
import sys
from dataclasses import asdict, dataclass, field
from typing import cast

import yaml
from PySide6.QtWidgets import QApplication

import chipiron as ch
import chipiron.displays as display
from chipiron.environments.chess.board import BoardFactory, create_board_factory
from chipiron.games.match.match_args import MatchArgs
from chipiron.games.match.match_factories import create_match_manager_from_args
from chipiron.scripts.chipiron_args import ImplementationArgs
from chipiron.scripts.script import Script
from chipiron.scripts.script_args import BaseScriptArgs
from chipiron.utils.dataclass import IsDataclass
from chipiron.utils.logger import chipiron_logger


@dataclass
class MatchScriptArgs:
    """
    The input arguments needed by the one match script to run
    """

    base_script_args: BaseScriptArgs = field(default_factory=BaseScriptArgs)
    match_args: MatchArgs = field(default_factory=MatchArgs)
    implementation_args: ImplementationArgs = field(default_factory=ImplementationArgs)
    # whether to display the match in a GUI
    gui: bool = True


class OneMatchScript:
    """
    Script that plays a match between two players

    """

    args_dataclass_name: type[MatchScriptArgs] = MatchScriptArgs

    base_experiment_output_folder = os.path.join(
        Script.base_experiment_output_folder, "one_match/outputs/"
    )
    base_script: Script[MatchScriptArgs]

    chess_gui: QApplication

    def __init__(
        self,
        base_script: Script[MatchScriptArgs],
    ) -> None:
        """
        Builds the OneMatchScript object
        """

        self.base_script = base_script

        # Calling the init of Script that takes care of a lot of stuff, especially parsing the arguments into args
        args: MatchScriptArgs = self.base_script.initiate(
            experiment_output_folder=self.base_experiment_output_folder,
        )

        # creating the match manager
        self.match_manager: ch.game.MatchManager = create_match_manager_from_args(
            match_args=args.match_args,
            base_script_args=args.base_script_args,
            implementation_args=args.implementation_args,
            gui=args.gui,
        )

        # saving the arguments of the script
        with open(
            os.path.join(
                str(args.base_script_args.experiment_output_folder),
                "inputs_and_parsing/one_match_script_merge.yaml",
            ),
            "w",
        ) as one_match_script:
            yaml.dump(asdict(args), one_match_script, default_flow_style=False)

        # checking for some incompatibility
        if args.gui and args.base_script_args.profiling:
            raise ValueError("Profiling does not work well atm with gui on")

        # If we need a GUI
        if args.gui:
            # if we use a graphic user interface (GUI) we create it its own thread and
            # create its mailbox to communicate with other threads
            gui_thread_mailbox: queue.Queue[IsDataclass] = (
                multiprocessing.Manager().Queue()
            )

            app = QApplication.instance()
            if app is None:
                app = QApplication(sys.argv)
            else:
                app = cast(QApplication, app)
            self.chess_gui = app

            board_factory: BoardFactory = create_board_factory(
                use_rust_boards=args.implementation_args.use_rust_boards,
                sort_legal_moves=args.base_script_args.universal_behavior,
            )
            self.window: display.MainWindow = display.MainWindow(
                gui_mailbox=gui_thread_mailbox,
                main_thread_mailbox=self.match_manager.game_manager_factory.main_thread_mailbox,
                board_factory=board_factory,
            )
            self.match_manager.subscribe(gui_thread_mailbox)

        self.gui = args.gui

    def run(self) -> None:
        """
        Runs the match either with a GUI or not
        Returns:

        """

        chipiron_logger.info(" Script One Match go")
        # Qt Application needs to be in the main Thread, so we need to distinguish between GUI and no GUI
        if self.gui:  # case with GUI
            # Launching the Match Manager in a Thread
            self.process_match_manager = multiprocessing.Process(
                target=self.match_manager.play_one_match
            )
            self.process_match_manager.start()

            # Qt Application launched in the main thread
            self.window.show()
            self.chess_gui.exec_()
        else:  # No GUI
            self.match_manager.play_one_match()

        chipiron_logger.info("Finish the run of the match")

        # TODO check the good closing of processes

    def terminate(self) -> None:
        """
        Terminates the script and cleans up any resources.
        """
        chipiron_logger.info("terminating script")
        self.base_script.terminate()
        if self.gui:
            self.process_match_manager.terminate()
